from __future__ import annotations

import time
import datetime

from typing import Optional, Any, Union, Dict, TYPE_CHECKING
from multiprocessing import Queue as MakeQueue
from multiprocessing.queues import Queue
from queue import Empty

from pibble.api.configuration import APIConfiguration
from pibble.util.strings import Serializer
from pibble.util.helpers import resolve

from enfugue.diffusion.process import DiffusionEngineProcess
from enfugue.util import logger

if TYPE_CHECKING:
    from enfugue.diffusion.plan import DiffusionPlan


__all__ = ["DiffusionEngine"]


class DiffusionEngine:
    LOGGING_DELAY_MS = 10000

    process: DiffusionEngineProcess

    def __init__(self, configuration: Optional[APIConfiguration] = None):
        self.configuration = APIConfiguration()
        self.request = 0

        if configuration is not None:
            self.configuration = configuration

    def __enter__(self) -> DiffusionEngine:
        """
        When entering the engine via context manager, start the process.
        """
        self.spawn_process(fail_when_redundant=False)
        return self

    def __exit__(self, *args: Any) -> None:
        """
        When exiting the engine via context manager, kill the process.
        """
        self.terminate_process()

    def check_get_queue(self, queue_name: str) -> Queue:
        """
        Gets a queue, creates it if not already made.
        """
        if not hasattr(self, f"_{queue_name}"):
            if self.is_alive():
                raise IOError(f"Queue '{queue_name}' is missing, but process is alive.")
            setattr(self, f"_{queue_name}", MakeQueue())
        return getattr(self, f"_{queue_name}")

    def check_delete_queue(self, queue_name: str) -> None:
        """
        Deletes a queue.
        """
        if hasattr(self, f"_{queue_name}"):
            if self.is_alive():
                raise IOError(f"Attempted to close queue {queue_name} while process is still alive")
            delattr(self, f"_{queue_name}")

    @property
    def instructions(self) -> Queue:
        """
        Gets the instructions queue
        """
        return self.check_get_queue("instructions")

    @instructions.deleter
    def instructions(self) -> None:
        """
        Deletes the instructions queue
        """
        self.check_delete_queue("instructions")

    @property
    def results(self) -> Queue:
        """
        Gets the results queue
        """
        return self.check_get_queue("results")

    @results.deleter
    def results(self) -> None:
        """
        Deletes the results queue
        """
        self.check_delete_queue("results")

    @property
    def intermediates(self) -> Queue:
        """
        Gets the intemerdiates queue
        """
        return self.check_get_queue("intermediates")

    @intermediates.deleter
    def intermediates(self) -> None:
        """
        Deletes the intermediates queue
        """
        self.check_delete_queue("intermediates")

    def spawn_process(
        self,
        timeout: Optional[Union[int, float]] = None,
        fail_when_redundant: bool = True,
    ) -> None:
        """
        Starts the engine process and waits for a response
        """
        if hasattr(self, "process") and self.process.is_alive():
            if fail_when_redundant:
                raise IOError("spawn_process called while a process is already running")
            return

        logger.debug("No current engine process, creating.")
        self.process = DiffusionEngineProcess(
            self.instructions, self.results, self.intermediates, self.configuration
        )

        poll_delay_seconds = DiffusionEngineProcess.POLLING_DELAY_MS / 250

        try:
            logger.debug("Starting process.")
            self.process.start()
            time.sleep(poll_delay_seconds)
            if not self.is_alive():
                raise IOError("Engine process died before it became responsive")
        except Exception as ex:
            self.terminate_process()
            raise ex

    def is_alive(self, timeout: Optional[Union[int, float]] = None) -> bool:
        """
        Determines if the process is alive.
        """
        if not hasattr(self, "process"):
            return False

        if not self.process.is_alive():
            self.terminate_process()
            return False

        return True

    def terminate_process(self, timeout: Optional[Union[int, float]] = 10) -> None:
        """
        Stops the process if it is running.
        """
        if not hasattr(self, "process"):
            return

        if self.process.is_alive():
            start = datetime.datetime.now()
            sleep_time = DiffusionEngineProcess.POLLING_DELAY_MS / 500
            self.dispatch("stop")
            time.sleep(sleep_time)
            while self.process.is_alive():
                waited = (datetime.datetime.now() - start).total_seconds()
                if timeout is not None and timeout > waited:
                    logger.debug("Process did not stop on it's own, sending TERM")
                    self.process.terminate()
                    time.sleep(sleep_time)
                    break
                time.sleep(sleep_time)
        if hasattr(self, "process") and self.process.is_alive():
            raise IOError("Couldn't terminate process")

        del self.process
        del self.intermediates
        del self.results
        del self.instructions

    def keepalive(self, timeout: Union[int, float] = 0.2) -> bool:
        """
        Keeps the process alive if it isn't already.
        """
        if not hasattr(self, "process"):
            return False

        if not self.process.is_alive():
            self.terminate_process()
            return False

        if self.instructions.empty():
            ping_response = self.invoke("ping")
            if ping_response != "pong":
                self.terminate_process()
                raise IOError(f"Incorrect ping response {ping_response}")
        return True

    def dispatch(self, action: str, payload: Any = None, spawn_process: bool = True) -> Any:
        """
        Sends a payload, does not wait for a response.
        """
        if spawn_process:
            self.spawn_process(timeout=1, fail_when_redundant=False)
        self.request += 1
        envelope = {"id": self.request, "action": action, "payload": payload}
        logger.debug(f"Dispatching envelope {envelope} to {self.instructions}")
        self.instructions.put(Serializer.serialize(envelope))
        return envelope["id"]

    def last_intermediate(self, id: int) -> Any:
        """
        Gets the last steps' details, if any
        """
        all_steps = []

        try:
            while True:
                all_steps.append(self.intermediates.get_nowait())
        except Empty:
            pass

        intermediate_data: Optional[Dict[str, Any]] = None
        for step in all_steps:
            step_deserialized = Serializer.deserialize(step)
            if step_deserialized["id"] == id:
                if intermediate_data is None:
                    intermediate_data = {"id": id}
                for key in ["step", "total", "rate", "images"]:
                    if key in step_deserialized:
                        intermediate_data[key] = step_deserialized[key]
            else:
                # Wrong ID, put back in queue
                self.intermediates.put_nowait(step)
        return intermediate_data

    def wait(self, id: int, timeout: Optional[Union[int, float]] = None) -> Any:
        """
        Waits for a response after issuing a request and getting an ID
        """
        start = datetime.datetime.now()
        poll_delay_seconds = DiffusionEngineProcess.POLLING_DELAY_MS / 500

        while True:
            to_raise, to_return = None, None
            all_results = []
            try:
                while True:
                    all_results.append(self.results.get_nowait())
            except Empty:
                pass

            for result in all_results:
                result_deserialized = Serializer.deserialize(result)
                try:
                    if result_deserialized["id"] == id:
                        if "error" in result_deserialized:
                            to_raise = resolve(result_deserialized["error"])(
                                result_deserialized["message"]
                            )
                            if "trace" in result_deserialized:
                                logger.error(result_deserialized["trace"])
                        else:
                            to_return = result_deserialized["result"]
                    else:
                        self.results.put_nowait(result)
                except:
                    logger.error(f"Couldn't parse result {result}")
                    raise
            if to_raise is not None:
                raise to_raise
            if to_return is not None:
                return to_return

            time.sleep(poll_delay_seconds)
            if timeout is not None and (datetime.datetime.now() - start).total_seconds() > timeout:
                raise TimeoutError("Timed out waiting for response.")

    def invoke(
        self, action: str, payload: Any = None, timeout: Optional[Union[int, float]] = None
    ) -> Any:
        """
        Issue a single request synchronously using arg syntax.
        """
        return self.wait(self.dispatch(action, payload), timeout)

    def execute(
        self, plan: DiffusionPlan, timeout: Optional[Union[int, float]] = None, wait: bool = False
    ) -> Any:
        """
        This is a helpful method to just serialize and execute a plan.
        """
        id = self.dispatch("plan", plan.get_serialization_dict())
        if wait:
            return self.wait(id, timeout)
        return id

    def __call__(self, timeout: Optional[Union[int, float]] = None, **kwargs: Any) -> Any:
        """
        Issues a single invocation request using kwarg syntax.
        """
        return self.wait(self.dispatch("invoke", kwargs), timeout)



import torch

from typing import List, Dict, Tuple

from onnx.onnx_ml_pb2 import ModelProto as ONNXModel

from enfugue.diffusion.rt.model.base import BaseModel
from enfugue.diffusion.rt.optimizer import Optimizer

__all__ = ["CLIP"]


class CLIP(BaseModel):
    """
    This is the model for CLiP (text embeddings)
    """

    def __init__(
        self,
        model: torch.nn.Module,
        use_fp16: bool = False,
        device: str = "cuda",
        max_batch_size: int = 16,
        embedding_dim: int = 768,
        text_maxlen: int = 77,
    ):
        super(CLIP, self).__init__(
            model,
            use_fp16=False,  # Override
            text_maxlen=77,  # Override
            device=device,
            max_batch_size=max_batch_size,
            embedding_dim=embedding_dim,
        )
        self.name = "CLIP"

    def get_model_key(self) -> str:
        """
        Gets the unique key for this model
        """
        return "clip"

    def get_input_names(self) -> List[str]:
        """
        Gets the input names that will be passed into the graph.
        """
        return ["input_ids"]

    def get_output_names(self) -> List[str]:
        """
        Gets the output names that will come from the graph.
        """
        return ["text_embeddings", "pooled_output"]

    def get_dynamic_axes(self) -> Dict[str, Dict[int, str]]:
        """
        Gets dynamic axes that can be shared.
        """
        return {"input_ids": {0: "B"}, "text_embeddings": {0: "B"}}

    def get_input_profile(
        self,
        batch_size: int,
        image_height: int,
        image_width: int,
        static_batch: bool,
        static_shape: bool,
    ) -> Dict[str, List[Tuple[int, ...]]]:
        """
        Gets the optimal dimensions for all inputs
        """
        self.check_dims(batch_size, image_height, image_height)
        (min_batch, max_batch), _, _, _, _ = self.get_minmax_dims(
            batch_size, image_height, image_width, static_batch, static_shape
        )
        return {
            "input_ids": [
                (min_batch, self.text_maxlen),
                (batch_size, self.text_maxlen),
                (max_batch, self.text_maxlen),
            ]
        }

    def get_shape_dict(
        self, batch_size: int, image_height: int, image_width: int
    ) -> Dict[str, Tuple[int, ...]]:
        self.check_dims(batch_size, image_height, image_width)
        return {
            "input_ids": (batch_size, self.text_maxlen),
            "text_embeddings": (batch_size, self.text_maxlen, self.embedding_dim),
        }

    def get_sample_input(
        self, batch_size: int, image_height: int, image_width: int
    ) -> torch.Tensor:
        self.check_dims(batch_size, image_height, image_width)
        return torch.zeros(batch_size, self.text_maxlen, dtype=torch.int32, device=self.device)

    def optimize(self, onnx_graph: ONNXModel) -> ONNXModel:
        return Optimizer.run(
            onnx_graph,
            select_inputs=[0],
            select_outputs=[0],
            select_output_names=["text_embeddings"],
        )



import os
import datetime
from typing import Optional, Dict, Callable, Tuple, Any, cast
from multiprocessing import Process, Queue as MakeQueue
from multiprocessing.queues import Queue
from queue import Empty
from enfugue.util import logger

__all__ = ["DownloadProcess", "Download"]


class DownloadProcess(Process):
    """
    A process to download a file to a location, optionally reporting progress.
    """

    def __init__(
        self,
        src: str,
        dest: str,
        chunk_size: int = 8192,
        headers: Dict[str, str] = {},
        parameters: Dict[str, Any] = {},
        progress: Optional[Queue] = None,
    ) -> None:
        super(DownloadProcess, self).__init__()
        self.src = src
        self.dest = dest
        self.chunk_size = chunk_size
        self.headers = headers
        self.parameters = parameters
        self.progress = progress

    def run(self) -> None:
        """
        Runs the download process, optionally sending progress over the queue.
        """
        from requests import get

        try:
            start = datetime.datetime.now()
            elapsed: Callable[[], float] = lambda: (datetime.datetime.now() - start).total_seconds()
            response = get(self.src, headers=self.headers, params=self.parameters, stream=True)
            length = int(response.headers["Content-Length"])
            if self.progress is not None:
                self.progress.put_nowait((elapsed(), 0, length))
            with open(self.dest, "wb") as fh:
                for i, chunk in enumerate(response.iter_content(chunk_size=self.chunk_size)):
                    fh.write(chunk)
                    if self.progress is not None:
                        self.progress.put_nowait((elapsed(), (i + 1) * self.chunk_size, length))
            if self.progress is not None:
                self.progress.put_nowait((elapsed(), length, length))
        except Exception as ex:
            logger.critical(f"Received critical error during download: {type(ex).__name__} {ex}")
            raise


class Download:
    """
    This class holds the queue and the process for downloading a file.
    """

    progress_queue: Optional[Queue]
    create_time: datetime.datetime
    start_time: Optional[datetime.datetime]
    close_time: Optional[datetime.datetime]
    last_elapsed: Optional[float]
    last_downloaded_bytes: Optional[int]
    last_total_bytes: Optional[int]

    def __init__(
        self,
        src: str,
        dest: str,
        chunk_size: int = 8192,
        headers: Dict[str, str] = {},
        parameters: Dict[str, Any] = {},
        progress: bool = True,
    ) -> None:
        self.src = src
        self.dest = dest
        self.create_time = datetime.datetime.now()
        self.progress_queue = None if not progress else MakeQueue()
        self.start_time = None
        self.close_time = None
        self.last_elapsed = None
        self.last_downloaded_bytes = None
        self.last_total_bytes = None
        self.process = DownloadProcess(
            src, dest, chunk_size, headers, parameters, self.progress_queue
        )

    def start(self) -> None:
        """
        Starts the process, and sets the start time.
        """
        if self.start_time is not None:
            raise IOError("Process already started.")
        logger.debug(f"Starting download process for {self.src}")
        self.start_time = datetime.datetime.now()
        self.process.start()

    def close(self) -> None:
        """
        Closes the queue and sets close time.
        """
        if self.closed:
            raise IOError("Already closed!")
        self.close_time = datetime.datetime.now()
        if self.progress_queue:
            self.progress_queue.close()
        if self.process.is_alive():
            logger.warning("Closed while process was still alive. Trying to terminate.")
            self.process.terminate()

    def get_last_progress(
        self,
    ) -> Optional[Tuple[Optional[float], Optional[int], Optional[int]]]:
        """
        Checks if there is new progress data from the queue, reads it all, then returns the last.
        """
        if not self.started:
            return None
        if self.progress_queue is None:
            raise ValueError(
                "Download was initialized with `progress=False`, no progress report is available."
            )
        if not self.closed:
            last_tuple = None
            try:
                while True:
                    last_tuple = self.progress_queue.get_nowait()
            except Empty:
                if not self.process.is_alive():
                    self.close()
            if last_tuple is not None:
                (
                    self.last_elapsed,
                    self.last_downloaded_bytes,
                    self.last_total_bytes,
                ) = last_tuple
        return self.last_elapsed, self.last_downloaded_bytes, self.last_total_bytes

    def check_raise_exitcode(self) -> None:
        """
        Checks if the process is alive, and if not, checks its exit code.
        If the process exited with an error, this will an IOError.
        """
        if self.started and not self.process.is_alive() and self.process.exitcode != 0:
            raise IOError(f"Process died with exit code {self.process.exitcode}")

    def format(self) -> Dict[str, Any]:
        """
        Gets the entire details of the download in one dictionary.
        """
        elapsed, downloaded, total = None, None, None
        status = "queued"
        if self.started:
            status = "complete" if self.closed else "downloading"
            last_progress = self.get_last_progress()
            if last_progress is not None:
                elapsed, downloaded, total = last_progress
            self.check_raise_exitcode()
        return {
            "status": status,
            "source": self.src,
            "destination": self.dest,
            "filename": os.path.basename(self.dest),
            "elapsed": elapsed,
            "downloaded": downloaded,
            "total": total,
        }

    @property
    def started(self) -> bool:
        """
        A quick helper for a boolean 'started' attribute
        """
        return self.start_time is not None

    @property
    def closed(self) -> bool:
        """
        Returns true if close time is set and queue is closed.
        """
        return self.close_time is not None

    @property
    def complete(self) -> bool:
        """
        Determines if the download is completed.
        """
        if not self.started:
            return False
        if self.closed:
            return True
        self.check_raise_exitcode()
        self.get_last_progress()
        return not self.process.is_alive()

    @property
    def progress(self) -> Optional[float]:
        """
        Gets the last progress from 0 to 1.0.
        Raises errors if the process failed.
        """
        if not self.started:
            return None
        self.check_raise_exitcode()
        self.get_last_progress()
        if self.last_downloaded_bytes is None or self.last_total_bytes is None:
            return None
        return self.last_downloaded_bytes / self.last_total_bytes

    @property
    def downloaded_bytes(self) -> Optional[int]:
        """
        Gets the last number of bytes downloaded.
        Raises errors if the process failed.
        """
        if not self.started:
            return None
        self.check_raise_exitcode()
        self.get_last_progress()
        return self.last_downloaded_bytes

    @property
    def total_bytes(self) -> Optional[int]:
        """
        Gets the last total number of bytes.
        Raises errors if the process failed.
        """
        if not self.started:
            return None
        self.check_raise_exitcode()
        self.get_last_progress()
        return self.last_total_bytes

    @property
    def elapsed(self) -> Optional[float]:
        """
        Gets the last total number of bytes.
        Raises errors if the process failed.
        """
        if not self.started:
            return None
        self.check_raise_exitcode()
        self.get_last_progress()
        if not self.last_elapsed:
            return (
                datetime.datetime.now() - cast(datetime.datetime, self.start_time)
            ).total_seconds()
        return self.last_elapsed



from __future__ import annotations

import cv2
import PIL
import numpy as np

__all__ = ["ComputerVision"]


class ComputerVision:
    """
    Provides helper methods for cv2
    """

    @staticmethod
    def convert_image(image: PIL.Image.Image) -> np.ndarray:
        """
        Converts PIL image to OpenCV format.
        """
        return cv2.cvtColor(np.array(image), cv2.COLOR_RGB2BGR)

    @staticmethod
    def revert_image(array: np.ndarray) -> PIL.Image.Image:
        """
        Converts PIL image to OpenCV format.
        """
        return PIL.Image.fromarray(cv2.cvtColor(array, cv2.COLOR_BGR2RGB))



import os
import requests

from enfugue.util.log import logger

__all__ = ["check_download", "check_download_to_dir"]


def check_download(remote_url: str, local_path: str, chunk_size: int = 8192) -> None:
    """
    Checks if a file exists.
    If it does, checks the size and matches against the remote URL.
    If it doesn't, or the size doesn't match, download it.
    """
    if os.path.exists(local_path):
        expected_length = requests.head(remote_url, allow_redirects=True).headers.get(
            "Content-Length", None
        )
        actual_length = os.path.getsize(local_path)
        if expected_length and actual_length != int(expected_length):
            logger.debug(
                f"File at {local_path} looks like an interrupted download, or the remote resource has changed. Removing."
            )
            os.remove(local_path)

    if not os.path.exists(local_path):
        logger.debug(f"Downloading file from {remote_url}. Will write to {local_path}")
        response = requests.get(remote_url, allow_redirects=True, stream=True)
        with open(local_path, "wb") as fh:
            for chunk in response.iter_content(chunk_size=chunk_size):
                fh.write(chunk)


def check_download_to_dir(remote_url: str, local_dir: str, chunk_size: int = 8192) -> str:
    """
    Checks if a file exists in a directory based on a remote path.
    If it does, checks the size and matches against the remote URL.
    If it doesn't, or the size doesn't match, download it.
    """
    file_name = os.path.basename(remote_url)
    local_path = os.path.join(local_dir, file_name)
    check_download(remote_url, local_path, chunk_size=chunk_size)
    return local_path



__all__ = [
    "DEFAULT_MODEL",
    "DEFAULT_INPAINTING_MODEL",
    "CONTROLNET_CANNY",
    "CONTROLNET_MLSD",
    "CONTROLNET_HED",
    "CONTROLNET_SCRIBBLE",
    "CONTROLNET_TILE",
    "CONTR0LNET_INPAINT",
]

DEFAULT_MODEL = (
    "https://huggingface.co/runwayml/stable-diffusion-v1-5/resolve/main/v1-5-pruned.ckpt"
)
DEFAULT_INPAINTING_MODEL = "https://huggingface.co/runwayml/stable-diffusion-inpainting/resolve/main/sd-v1-5-inpainting.ckpt"

CONTROLNET_CANNY = "lllyasviel/sd-controlnet-canny"
CONTROLNET_MLSD = "lllyasviel/control_v11p_sd15_mlsd"
CONTROLNET_HED = "lllyasviel/sd-controlnet-hed"
CONTROLNET_SCRIBBLE = "lllyasviel/control_v11p_sd15_scribble"
CONTROLNET_TILE = "lllyasviel/control_v11f1e_sd15_tile"
CONTROLNET_INPAINT = "lllyasviel/control_v11p_sd15_inpaint"



from __future__ import annotations

import PIL

from typing import Union

from realesrgan import RealESRGANer
from basicsr.archs.rrdbnet_arch import RRDBNet

from enfugue.util import check_download_to_dir
from enfugue.diffusion.vision import ComputerVision


class Upscaler:
    """
    The upscaler user ESRGAN or GFGPGAN for up to 4x upscale
    """

    GFPGAN_PATH = "https://github.com/TencentARC/GFPGAN/releases/download/v1.3.0/GFPGANv1.3.pth"
    ESRGAN_PATH = (
        "https://github.com/xinntao/Real-ESRGAN/releases/download/v0.1.0/RealESRGAN_x4plus.pth"
    )
    ESRGAN_ANIME_PATH = "https://github.com/xinntao/Real-ESRGAN/releases/download/v0.2.2.4/RealESRGAN_x4plus_anime_6B.pth"

    def __init__(self, model_dir: str) -> None:
        """
        On initialization, pass the model dir.
        """
        self.model_dir = model_dir

    @property
    def esrgan_weights_path(self) -> str:
        """
        Gets the path to the ESRGAN model weights.
        """
        return check_download_to_dir(self.ESRGAN_PATH, self.model_dir)

    @property
    def esrgan_anime_weights_path(self) -> str:
        """
        Gets the path to the ESRGAN anime model weights.
        """
        return check_download_to_dir(self.ESRGAN_ANIME_PATH, self.model_dir)

    @property
    def gfpgan_weights_path(self) -> str:
        """
        Gets the path to the GFPGAN model weights.
        """
        return check_download_to_dir(self.GFPGAN_PATH, self.model_dir)

    def get_upsampler(
        self, tile: int = 0, tile_pad: int = 10, pre_pad: int = 10, anime: bool = False
    ) -> RealESRGANer:
        """
        Gets the appropriate upsampler
        """
        if anime:
            model = RRDBNet(
                num_in_ch=3, num_out_ch=3, num_feat=64, num_block=6, num_grow_ch=32, scale=4
            )
            model_path = self.esrgan_anime_weights_path
        else:
            model = RRDBNet(
                num_in_ch=3, num_out_ch=3, num_feat=64, num_block=23, num_grow_ch=32, scale=4
            )
            model_path = self.esrgan_weights_path

        return RealESRGANer(
            scale=4,
            model_path=model_path,
            model=model,
            tile=tile,
            tile_pad=tile_pad,
            pre_pad=pre_pad,
            half=True,
        )

    def esrgan(
        self,
        image: Union[str, PIL.Image.Image],
        outscale: int = 4,
        tile: int = 0,
        tile_pad: int = 10,
        pre_pad: int = 10,
        anime: bool = False,
    ) -> PIL.Image.Image:
        """
        Does a simple upscale
        """
        if type(image) is str:
            image = PIL.Image.open(image)

        esrganer = self.get_upsampler(tile=tile, tile_pad=tile_pad, pre_pad=pre_pad, anime=anime)

        return ComputerVision.revert_image(
            esrganer.enhance(ComputerVision.convert_image(image), outscale=outscale)[0]
        )

    def gfpgan(
        self,
        image: Union[str, PIL.Image.Image],
        outscale: int = 4,
        tile: int = 0,
        tile_pad: int = 10,
        pre_pad: int = 10,
    ) -> PIL.Image.Image:
        """
        Does an upscale with face enhancement
        """
        if type(image) is str:
            image = PIL.Image.open(image)

        from enfugue.diffusion.upscale.gfpganer import GFPGANer  # type: ignore[attr-defined]

        gfpganer = GFPGANer(
            model_path=self.gfpgan_weights_path,
            upscale=outscale,
            arch="clean",
            channel_multiplier=2,
            bg_upsampler=self.get_upsampler(tile=tile, tile_pad=tile_pad, pre_pad=pre_pad),
            rootpath=self.model_dir,
        )

        return ComputerVision.revert_image(
            gfpganer.enhance(
                ComputerVision.convert_image(image),
                has_aligned=False,
                only_center_face=False,
                paste_back=True,
            )[2]
        )



# type: ignore
# adapted from https://raw.githubusercontent.com/anderskm/gputil/master/GPUtil/GPUtil.py

import os
import platform
from subprocess import Popen, PIPE
from distutils import spawn

try:
    from subprocess import CREATE_NO_WINDOW

    PROCESS_KWARGS = {"creationflags": CREATE_NO_WINDOW}
except ImportError:
    PROCESS_KWARGS = {}


class GPU:
    def __init__(
        self,
        ID,
        uuid,
        load,
        memoryTotal,
        memoryUsed,
        memoryFree,
        driver,
        gpu_name,
        serial,
        display_mode,
        display_active,
        temp_gpu,
    ):
        self.id = ID
        self.uuid = uuid
        self.load = load
        self.memoryUtil = float(memoryUsed) / float(memoryTotal)
        self.memoryTotal = memoryTotal
        self.memoryUsed = memoryUsed
        self.memoryFree = memoryFree
        self.driver = driver
        self.name = gpu_name
        self.serial = serial
        self.display_mode = display_mode
        self.display_active = display_active
        self.temperature = temp_gpu


def safeFloatCast(strNumber):
    try:
        number = float(strNumber)
    except ValueError:
        number = float("nan")
    return number


def getGPUs():
    if platform.system() == "Windows":
        # If the platform is Windows and nvidia-smi
        # could not be found from the environment path,
        # try to find it from system drive with default installation path
        nvidia_smi = spawn.find_executable("nvidia-smi")
        if nvidia_smi is None:
            nvidia_smi = (
                "%s\\Program Files\\NVIDIA Corporation\\NVSMI\\nvidia-smi.exe"
                % os.environ["systemdrive"]
            )
    else:
        nvidia_smi = "nvidia-smi"

    # Get ID, processing and memory utilization for all GPUs
    try:
        p = Popen(
            [
                nvidia_smi,
                "--query-gpu=index,uuid,utilization.gpu,memory.total,memory.used,memory.free,driver_version,name,gpu_serial,display_active,display_mode,temperature.gpu",
                "--format=csv,noheader,nounits",
            ],
            stdout=PIPE,
            stderr=PIPE,
            **PROCESS_KWARGS
        )
        stdout, stderror = p.communicate()
    except:
        return []
    output = stdout.decode("UTF-8")
    # output = output[2:-1] # Remove b' and ' from string added by python
    # print(output)
    ## Parse output
    # Split on line break
    lines = output.split(os.linesep)
    # print(lines)
    numDevices = len(lines) - 1
    GPUs = []
    for g in range(numDevices):
        line = lines[g]
        # print(line)
        vals = line.split(", ")
        # print(vals)
        for i in range(12):
            # print(vals[i])
            if i == 0:
                deviceIds = int(vals[i])
            elif i == 1:
                uuid = vals[i]
            elif i == 2:
                gpuUtil = safeFloatCast(vals[i]) / 100
            elif i == 3:
                memTotal = safeFloatCast(vals[i])
            elif i == 4:
                memUsed = safeFloatCast(vals[i])
            elif i == 5:
                memFree = safeFloatCast(vals[i])
            elif i == 6:
                driver = vals[i]
            elif i == 7:
                gpu_name = vals[i]
            elif i == 8:
                serial = vals[i]
            elif i == 9:
                display_active = vals[i]
            elif i == 10:
                display_mode = vals[i]
            elif i == 11:
                temp_gpu = safeFloatCast(vals[i])
        GPUs.append(
            GPU(
                deviceIds,
                uuid,
                gpuUtil,
                memTotal,
                memUsed,
                memFree,
                driver,
                gpu_name,
                serial,
                display_mode,
                display_active,
                temp_gpu,
            )
        )
    return GPUs  # (deviceIds, gpuUtil, memUtil)



import torch
from typing import Tuple, Dict, List

from enfugue.diffusion.rt.model.base import BaseModel


class VAE(BaseModel):
    def __init__(
        self,
        model: torch.nn.Module,
        device: str,
        max_batch_size: int,
        embedding_dim: int,
        use_fp16: bool,
    ):
        super(VAE, self).__init__(
            model=model,
            device=device,
            max_batch_size=max_batch_size,
            embedding_dim=embedding_dim,
        )
        self.name = "VAEDecoder"
        self.use_fp16 = use_fp16

    def get_model_key(self) -> str:
        return "vae"

    def get_input_names(self) -> List[str]:
        return ["latent"]

    def get_output_names(self) -> List[str]:
        return ["images"]

    def get_dynamic_axes(self) -> Dict[str, Dict[int, str]]:
        return {
            "latent": {0: "B", 2: "H", 3: "W"},
            "images": {0: "B", 2: "8H", 3: "8W"},
        }

    def get_input_profile(
        self,
        batch_size: int,
        image_height: int,
        image_width: int,
        static_batch: bool,
        static_shape: bool,
    ) -> Dict[str, List[Tuple[int, ...]]]:
        latent_height, latent_width = self.check_dims(batch_size, image_height, image_width)
        (
            (min_batch, max_batch),
            _,
            _,
            (min_latent_height, max_latent_height),
            (min_latent_width, max_latent_width),
        ) = self.get_minmax_dims(batch_size, image_height, image_width, static_batch, static_shape)
        return {
            "latent": [
                (min_batch, 4, min_latent_height, min_latent_width),
                (batch_size, 4, latent_height, latent_width),
                (max_batch, 4, max_latent_height, max_latent_width),
            ]
        }

    def get_shape_dict(
        self, batch_size: int, image_height: int, image_width: int
    ) -> Dict[str, Tuple[int, ...]]:
        latent_height, latent_width = self.check_dims(batch_size, image_height, image_width)
        return {
            "latent": (batch_size, 4, latent_height, latent_width),
            "images": (batch_size, 3, image_height, image_width),
        }

    def get_sample_input(
        self, batch_size: int, image_height: int, image_width: int
    ) -> torch.Tensor:
        latent_height, latent_width = self.check_dims(batch_size, image_height, image_width)
        return torch.randn(
            batch_size,
            4,
            latent_height,
            latent_width,
            dtype=torch.float32,
            device=self.device,
        )



from __future__ import annotations

import math
import PIL
import PIL.Image
import PIL.ImageDraw
import PIL.ImageOps

from typing import (
    Optional,
    Dict,
    Any,
    Union,
    Tuple,
    List,
    Literal,
    Callable,
    TypedDict,
    TYPE_CHECKING,
)

from enfugue.util import logger, feather_mask, fit_image, remove_background, TokenMerger

if TYPE_CHECKING:
    from enfugue.diffusers.manager import DiffusionPipelineManager
    from diffusers.pipelines.stable_diffusion import StableDiffusionPipelineOutput

DEFAULT_SIZE = 512
DEFAULT_IMAGE_CALLBACK_STEPS = 10
DEFAULT_CONDITIONING_SCALE = 1.0
DEFAULT_CHUNKING_SIZE = 64
DEFAULT_CHUNKING_BLUR = 64
DEFAULT_IMG2IMG_STRENGTH = 0.8
DEFAULT_INFERENCE_STEPS = 50
DEFAULT_GUIDANCE_SCALE = 7.5
DEFAULT_UPSCALE_PROMPT = (
    "highly detailed, ultra-detailed, intricate detail, high definition, HD, 4k, ultra high def"
)
DEFAULT_UPSCALE_NEGATIVE_PROMPT = "blurry, out-of-focus, ugly, pixelated, artifacts"
DEFAULT_UPSCALE_DIFFUSION_STEPS = 100
DEFAULT_UPSCALE_DIFFUSION_GUIDANCE_SCALE = 12
DEFAULT_UPSCALE_DIFFUSION_STRENGTH = 0.2
DEFAULT_UPSCALE_DIFFUSION_CHUNKING_SIZE = 64
DEFAULT_UPSCALE_DIFFUSION_CHUNKING_BLUR = 64

MODEL_PROMPT_WEIGHT = 0.2
GLOBAL_PROMPT_STEP_WEIGHT = 0.4
GLOBAL_PROMPT_UPSCALE_WEIGHT = 0.4
UPSCALE_PROMPT_STEP_WEIGHT = 0.1
MAX_IMAGE_SCALE = 3.0

__all__ = ["DiffusionStep", "DiffusionPlan"]


class NodeDict(TypedDict):
    w: int
    h: int
    x: int
    y: int
    fit: Optional[Literal["actual", "stretch", "contain", "cover"]]
    anchor: Optional[
        Literal[
            "top-left",
            "top-center",
            "top-right",
            "center-left",
            "center-center",
            "center-right",
            "bottom-left",
            "bottom-center",
            "bottom-right",
        ]
    ]
    infer: Optional[bool]
    inpaint: Optional[bool]
    control: Optional[bool]
    controlnet: Optional[str]
    prompt: Optional[str]
    negative_prompt: Optional[str]
    strength: Optional[float]
    conditioning_scale: Optional[float]
    image: Optional[PIL.Image.Image]
    mask: Optional[PIL.Image.Image]


class DiffusionStep:
    """
    A step represents most of the inputs to describe what the image is and how to control inference
    """

    result: StableDiffusionPipelineOutput

    def __init__(
        self,
        width: Optional[int] = None,
        height: Optional[int] = None,
        prompt: Optional[str] = None,
        negative_prompt: Optional[str] = None,
        image: Optional[Union[DiffusionStep, PIL.Image.Image, str]] = None,
        mask: Optional[Union[DiffusionStep, PIL.Image.Image, str]] = None,
        control_image: Optional[Union[DiffusionStep, PIL.Image.Image, str]] = None,
        controlnet: Optional[Literal["canny", "tile", "mlsd", "hed", "scribble", "inpaint"]] = None,
        conditioning_scale: Optional[float] = DEFAULT_CONDITIONING_SCALE,
        strength: Optional[float] = DEFAULT_IMG2IMG_STRENGTH,
        num_inference_steps: Optional[int] = DEFAULT_INFERENCE_STEPS,
        guidance_scale: Optional[float] = DEFAULT_GUIDANCE_SCALE,
        remove_background: bool = False,
        process_control_image: bool = True,
        scale_to_model_size: bool = True,
    ) -> None:
        self.width = width
        self.height = height
        self.prompt = prompt
        self.negative_prompt = negative_prompt
        self.image = image
        self.mask = mask
        self.control_image = control_image
        self.controlnet = controlnet
        self.strength = strength if strength is not None else DEFAULT_IMG2IMG_STRENGTH
        self.conditioning_scale = (
            conditioning_scale if conditioning_scale is not None else DEFAULT_CONDITIONING_SCALE
        )
        self.num_inference_steps = (
            num_inference_steps if num_inference_steps is not None else DEFAULT_INFERENCE_STEPS
        )
        self.guidance_scale = (
            guidance_scale if guidance_scale is not None else DEFAULT_GUIDANCE_SCALE
        )
        self.remove_background = remove_background
        self.process_control_image = process_control_image
        self.scale_to_model_size = scale_to_model_size

    def get_serialization_dict(self) -> Dict[str, Any]:
        """
        Gets the dictionary that will be returned to serialize
        """
        serialized: Dict[str, Any] = {
            "width": self.width,
            "height": self.height,
            "prompt": self.prompt,
            "negative_prompt": self.negative_prompt,
            "controlnet": self.controlnet,
            "conditioning_scale": self.conditioning_scale,
            "strength": self.strength,
            "num_inference_steps": self.num_inference_steps,
            "guidance_scale": self.guidance_scale,
            "remove_background": self.remove_background,
            "process_control_image": self.process_control_image,
            "scale_to_model_size": self.scale_to_model_size,
        }

        serialize_children: List[DiffusionStep] = []
        for key in ["image", "mask", "control_image"]:
            child = getattr(self, key)
            if isinstance(child, DiffusionStep):
                if child in serialize_children:
                    serialized[key] = serialize_children.index(child)
                else:
                    serialize_children.append(child)
                    serialized[key] = len(serialize_children) - 1
            else:
                serialized[key] = child

        serialized["children"] = [child.get_serialization_dict() for child in serialize_children]
        return serialized

    @property
    def kwargs(self) -> Dict[str, Any]:
        """
        Returns the keyword arguments that will passed to the pipeline invocation.
        """
        return {
            "width": self.width,
            "height": self.height,
            "prompt": self.prompt,
            "negative_prompt": self.negative_prompt,
            "image": self.image,
            "control_image": self.control_image,
            "conditioning_scale": self.conditioning_scale,
            "strength": self.strength,
            "num_inference_steps": self.num_inference_steps,
            "guidance_scale": self.guidance_scale,
        }

    def check_process_control_image(
        self, pipeline: DiffusionPipelineManager, control_image: Optional[PIL.Image.Image]
    ) -> Optional[PIL.Image.Image]:
        """
        Gets the control image for the pipeline based on the requested controlnet
        """
        if self.controlnet is None or control_image is None:
            return None
        if self.process_control_image:
            if self.controlnet == "canny":
                return pipeline.edge_detector.canny(control_image)
            if self.controlnet == "mlsd":
                return pipeline.edge_detector.mlsd(control_image)
            if self.controlnet == "hed":
                return pipeline.edge_detector.hed(control_image)
        return control_image

    def execute(
        self,
        pipeline: DiffusionPipelineManager,
        **kwargs: Any,
    ) -> StableDiffusionPipelineOutput:
        """
        Executes this pipeline step.
        """
        if hasattr(self, "result"):
            return self.result

        samples = kwargs.pop("samples", 1)

        if isinstance(self.image, DiffusionStep):
            image = self.image.execute(pipeline, samples=1, **kwargs)["images"][0]
        elif isinstance(self.image, str):
            image = PIL.Image.open(self.image)
        else:
            image = self.image

        if isinstance(self.mask, DiffusionStep):
            mask = self.mask.execute(pipeline, samples=1, **kwargs)["images"][0]
        elif isinstance(self.mask, str):
            mask = PIL.Image.open(self.mask)
        else:
            mask = self.mask

        if isinstance(self.control_image, DiffusionStep):
            control_image = self.control_image.execute(pipeline, samples=1, **kwargs)["images"][0]
        elif isinstance(self.control_image, str):
            control_image = PIL.Image.open(self.control_image)
        else:
            control_image = self.control_image

        if not self.prompt:
            if image:
                samples = kwargs.get("samples", 1)
                from diffusers.pipelines.stable_diffusion import StableDiffusionPipelineOutput

                self.result = StableDiffusionPipelineOutput(
                    images=[image] * samples, nsfw_content_detected=[False] * samples
                )
                return self.result
            raise ValueError("No prompt or image in this step; cannot invoke or pass through.")

        pipeline.controlnet = self.controlnet
        invocation_kwargs = {**kwargs, **self.kwargs}

        image_scale = 1
        image_width, image_height = None, None

        if image is not None:
            image_width, image_height = image.size
            invocation_kwargs["image"] = image
        if control_image is not None:
            if image is not None:
                assert image.size == control_image.size
            else:
                image_width, image_height = control_image.size
            invocation_kwargs["control_image"] = control_image
        if mask is not None:
            invocation_kwargs["mask"] = mask
            if image is not None:
                assert image.size == mask.size
            if control_image is not None:
                assert control_image.size == mask.size
            else:
                image_width, image_height = mask.size
        if (
            self.width is not None
            and self.height is not None
            and image_width is None
            and image_height is None
        ):
            image_width, image_height = self.width, self.height

        if image_width is None or image_height is None:
            raise ValueError("No known invocation size.")

        if image_width is not None and image_width < pipeline.size:
            image_scale = pipeline.size / image_width
        if image_height is not None and image_height < pipeline.size:
            image_scale = max(image_scale, pipeline.size / image_height)

        if image_scale > MAX_IMAGE_SCALE or not self.scale_to_model_size:
            # Refuse it's too oblong. We'll just calculate at the appropriate size.
            image_scale = 1

        invocation_kwargs["width"] = 8 * round((image_width * image_scale) / 8)
        invocation_kwargs["height"] = 8 * round((image_height * image_scale) / 8)
        invocation_kwargs["control_image"] = self.check_process_control_image(
            pipeline, control_image
        )

        if image_scale > 1:
            # scale input images up
            for key in ["image", "mask", "control_image"]:
                if invocation_kwargs.get(key, None) is not None:
                    invocation_kwargs[key] = self.scale_image(invocation_kwargs[key], image_scale)

        result = pipeline(**invocation_kwargs)

        if self.remove_background:
            for i, image in enumerate(result["images"]):
                result["images"][i] = remove_background(image)
                logger.critical(result["images"][i])

        if image_scale > 1:
            for i, image in enumerate(result["images"]):
                result["images"][i] = self.scale_image(image, 1 / image_scale)
                logger.critical(result["images"][i])

        self.result = result
        return result

    @staticmethod
    def scale_image(image: PIL.Image.Image, scale: Union[int, float]) -> PIL.Image.Image:
        """
        Scales an image proportionally.
        """
        width, height = image.size
        scaled_width = 8 * round((width * scale) / 8)
        scaled_height = 8 * round((height * scale) / 8)
        return image.resize((scaled_width, scaled_height))

    @staticmethod
    def deserialize_dict(step_dict: Dict[str, Any]) -> DiffusionStep:
        """
        Given a serialized dict, instantiate a diffusion step
        """
        kwargs: Dict[str, Any] = {}
        for key in [
            "prompt",
            "negative_prompt",
            "controlnet",
            "conditioning_scale",
            "strength",
            "num_inference_steps",
            "guidance_scale",
            "width",
            "height",
            "remove_background",
            "process_control_image",
            "scale_to_model_size",
        ]:
            if key in step_dict:
                kwargs[key] = step_dict[key]

        deserialized_children = [
            DiffusionStep.deserialize_dict(child) for child in step_dict.get("children", [])
        ]
        for key in ["image", "control_image", "mask"]:
            if key not in step_dict:
                continue
            if isinstance(step_dict[key], int):
                kwargs[key] = deserialized_children[step_dict[key]]
            else:
                kwargs[key] = step_dict[key]

        return DiffusionStep(**kwargs)


class DiffusionNode:
    """
    A diffusion node has a step that may be recursive, combined with bounds.
    """

    def __init__(self, bounds: List[Tuple[int, int]], step: DiffusionStep) -> None:
        self.bounds = bounds
        self.step = step

    def resize_image(self, image: PIL.Image.Image) -> PIL.Image.Image:
        """
        Resizes the image to fit the bounds.
        """
        x, y = self.bounds[0]
        w, h = self.bounds[1]
        return image.resize((w - x, h - y))

    def get_serialization_dict(self) -> Dict[str, Any]:
        """
        Gets the step's dict and adds bounds.
        """
        step_dict = self.step.get_serialization_dict()
        step_dict["bounds"] = self.bounds
        return step_dict

    def execute(
        self,
        pipeline: DiffusionPipelineManager,
        **kwargs: Any,
    ) -> StableDiffusionPipelineOutput:
        """
        Passes through the execution to the step.
        """
        return self.step.execute(pipeline, **kwargs)

    @staticmethod
    def deserialize_dict(step_dict: Dict[str, Any]) -> DiffusionNode:
        """
        Given a serialized dict, instantiate a diffusion Node
        """
        bounds = step_dict.pop("bounds", None)
        if bounds is None:
            raise TypeError("Bounds are required")

        return DiffusionNode(
            [(int(bounds[0][0]), int(bounds[0][1])), (int(bounds[1][0]), int(bounds[1][1]))],
            DiffusionStep.deserialize_dict(step_dict),
        )


class DiffusionPlan:
    """
    A diffusion plan represents any number of steps, with each step receiving the output of the previous.

    Additionally, we handle upscaling as part of the plan. If we want to upscale later, the Plan can be initiated
    with an empty steps array and initial image.
    """

    def __init__(
        self,
        prompt: Optional[str] = None,  # Global
        negative_prompt: Optional[str] = None,  # Global
        size: Optional[int] = DEFAULT_SIZE,
        model: Optional[str] = None,
        lora: Optional[
            Union[str, List[str], Tuple[str, float], List[Union[str, Tuple[str, float]]]]
        ] = None,
        inversion: Optional[Union[str, List[str]]] = None,
        width: Optional[int] = DEFAULT_SIZE,
        height: Optional[int] = DEFAULT_SIZE,
        image_callback_steps: Optional[int] = DEFAULT_IMAGE_CALLBACK_STEPS,
        nodes: List[DiffusionNode] = [],
        image: Optional[Union[str, PIL.Image.Image]] = None,
        chunking_size: Optional[int] = DEFAULT_CHUNKING_SIZE,
        chunking_blur: Optional[int] = DEFAULT_CHUNKING_BLUR,
        samples: Optional[int] = 1,
        seed: Optional[int] = None,
        build_tensorrt: bool = False,
        outscale: Optional[int] = 1,
        upscale: Optional[
            Union[
                Literal[
                    "esrgan",
                    "esrganime",
                    "gfpgan",
                    "lanczos",
                    "bilinear",
                    "bicubic",
                    "lanczos",
                    "nearest",
                ],
                List[
                    Literal[
                        "esrgan",
                        "esrganime",
                        "gfpgan",
                        "lanczos",
                        "bilinear",
                        "bicubic",
                        "lanczos",
                        "nearest",
                    ]
                ],
            ]
        ] = None,
        upscale_diffusion: bool = False,
        upscale_iterative: bool = False,
        upscale_diffusion_steps: Optional[Union[int, List[int]]] = DEFAULT_UPSCALE_DIFFUSION_STEPS,
        upscale_diffusion_guidance_scale: Optional[
            Union[float, int, List[Union[float, int]]]
        ] = DEFAULT_UPSCALE_DIFFUSION_GUIDANCE_SCALE,
        upscale_diffusion_strength: Optional[
            Union[float, List[float]]
        ] = DEFAULT_UPSCALE_DIFFUSION_STRENGTH,
        upscale_diffusion_prompt: Optional[Union[str, List[str]]] = DEFAULT_UPSCALE_PROMPT,
        upscale_diffusion_negative_prompt: Optional[
            Union[str, List[str]]
        ] = DEFAULT_UPSCALE_NEGATIVE_PROMPT,
        upscale_diffusion_controlnet: Optional[
            Union[
                Literal["canny", "tile", "mlsd", "hed", "scribble", "inpaint"],
                List[Literal["canny", "tile", "mlsd", "hed", "scribble", "inpaint"]],
            ]
        ] = None,
        upscale_diffusion_chunking_size: Optional[int] = DEFAULT_UPSCALE_DIFFUSION_CHUNKING_SIZE,
        upscale_diffusion_chunking_blur: Optional[int] = DEFAULT_UPSCALE_DIFFUSION_CHUNKING_BLUR,
        upscale_diffusion_scale_chunking_size: bool = True,
        upscale_diffusion_scale_chunking_blur: bool = True,
    ) -> None:
        self.size = size if size is not None else DEFAULT_SIZE
        self.prompt = prompt
        self.negative_prompt = negative_prompt
        self.model = model
        self.lora = lora
        self.inversion = inversion
        self.width = width if width is not None else DEFAULT_SIZE
        self.height = height if height is not None else DEFAULT_SIZE
        self.image = image
        self.image_callback_steps = image_callback_steps
        self.chunking_size = (
            chunking_size if chunking_size is not None else DEFAULT_CHUNKING_SIZE
        )  # Pass 0 to disable
        self.chunking_blur = (
            chunking_blur if chunking_blur is not None else DEFAULT_CHUNKING_BLUR
        )  # Pass 0 to disable
        self.samples = samples if samples is not None else 1
        self.seed = seed
        self.build_tensorrt = build_tensorrt
        self.nodes = nodes
        self.outscale = outscale if outscale is not None else 1
        self.upscale = upscale
        self.upscale_iterative = bool(upscale_iterative)
        self.upscale_diffusion = bool(upscale_diffusion)
        self.upscale_diffusion_chunking_size = (
            upscale_diffusion_chunking_size
            if upscale_diffusion_chunking_size is not None
            else DEFAULT_UPSCALE_DIFFUSION_CHUNKING_SIZE
        )
        self.upscale_diffusion_chunking_blur = (
            upscale_diffusion_chunking_blur
            if upscale_diffusion_chunking_blur is not None
            else DEFAULT_UPSCALE_DIFFUSION_CHUNKING_BLUR
        )
        self.upscale_diffusion_guidance_scale = (
            upscale_diffusion_guidance_scale
            if upscale_diffusion_guidance_scale is not None
            else DEFAULT_UPSCALE_DIFFUSION_GUIDANCE_SCALE
        )
        self.upscale_diffusion_steps = (
            upscale_diffusion_steps
            if upscale_diffusion_steps is not None
            else DEFAULT_UPSCALE_DIFFUSION_STEPS
        )
        self.upscale_diffusion_strength = (
            upscale_diffusion_strength
            if upscale_diffusion_strength is not None
            else DEFAULT_UPSCALE_DIFFUSION_STRENGTH
        )
        self.upscale_diffusion_prompt = (
            upscale_diffusion_prompt
            if upscale_diffusion_prompt is not None
            else DEFAULT_UPSCALE_PROMPT
        )
        self.upscale_diffusion_negative_prompt = (
            upscale_diffusion_negative_prompt
            if upscale_diffusion_negative_prompt is not None
            else DEFAULT_UPSCALE_NEGATIVE_PROMPT
        )
        self.upscale_diffusion_controlnet = upscale_diffusion_controlnet
        self.upscale_diffusion_scale_chunking_size = bool(upscale_diffusion_scale_chunking_size)
        self.upscale_diffusion_scale_chunking_blur = bool(upscale_diffusion_scale_chunking_blur)

    @property
    def kwargs(self) -> Dict[str, Any]:
        """
        Returns the keyword arguments that will be passing to the pipeline call.
        """
        return {
            "width": self.width,
            "height": self.height,
            "chunking_size": self.chunking_size,
            "chunking_blur": self.chunking_blur,
            "num_images_per_prompt": self.samples,
        }

    def execute(
        self,
        pipeline: DiffusionPipelineManager,
        progress_callback: Optional[Callable[[int, int, float], None]] = None,
        image_callback: Optional[Callable[[List[PIL.Image.Image]], None]] = None,
        image_callback_steps: Optional[int] = None,
    ) -> StableDiffusionPipelineOutput:
        """
        This is the main interface for execution.

        The first step will be the one that executes with the selected number of samples,
        and then each subsequent step will be performed on the number of outputs from the
        first step.
        """
        # We import here so this file can be imported by processes without initializing torch
        from diffusers.utils.pil_utils import PIL_INTERPOLATION
        from diffusers.pipelines.stable_diffusion import StableDiffusionPipelineOutput

        images, nsfw = self.execute_nodes(
            pipeline, progress_callback, image_callback, image_callback_steps
        )
        if self.upscale is not None and self.outscale > 1:
            if self.upscale_iterative:
                scales = [2] * int(math.log(self.outscale, 2))
            else:
                scales = [self.outscale]
            for j, scale in enumerate(scales):
                # Unload the pipeline to force buffers to refresh, saves VRAM but takes time
                pipeline.unload_pipeline()

                def get_item_for_scale(item: Any) -> Any:
                    if not isinstance(item, list):
                        return item
                    elif len(item) < j + 1:
                        if len(item) == 0:
                            return None
                        return item[-1]
                    else:
                        return item[j]

                for i, image in enumerate(images):
                    if nsfw is not None and nsfw[i]:
                        logger.debug(f"Image {i} had NSFW content, not upscaling.")
                        continue

                    upscale = get_item_for_scale(self.upscale)
                    logger.debug(f"Upscaling sample {i} by {scale} using {upscale}")

                    if upscale == "esrgan":
                        image = pipeline.upscaler.esrgan(image, tile=pipeline.size, outscale=scale)
                    elif upscale == "esrganime":
                        image = pipeline.upscaler.esrgan(
                            image, tile=pipeline.size, outscale=scale, anime=True
                        )
                    elif upscale == "gfpgan":
                        image = pipeline.upscaler.gfpgan(image, tile=pipeline.size, outscale=scale)
                    elif upscale in PIL_INTERPOLATION:
                        width, height = image.size
                        image = image.resize(
                            (width * scale, height * scale), resample=PIL_INTERPOLATION[upscale]
                        )
                    else:
                        logger.error(f"Unknown upscaler {upscale}")
                        return images
                    images[i] = image

                # Unload the upscaler to save VRAM
                pipeline.unload_upscaler()

                if image_callback is not None and (j < len(scales) - 1 or self.upscale_diffusion):
                    image_callback(images)

                if self.upscale_diffusion:
                    # Disable pipeline safety here, it gives many false positives when upscaling.
                    # We'll re-enable it after.
                    re_enable_safety = pipeline.safe
                    pipeline.safe = False
                    for i, image in enumerate(images):
                        if nsfw is not None and nsfw[i]:
                            logger.debug(f"Image {i} had NSFW content, not upscaling.")
                            continue
                        # Prepare the pipeline again
                        self.prepare_pipeline(pipeline)
                        width, height = image.size
                        kwargs = {
                            "width": width,
                            "height": height,
                            "image": image,
                            "num_images_per_prompt": 1,
                            "prompt": get_item_for_scale(self.upscale_diffusion_prompt),
                            "negative_prompt": get_item_for_scale(
                                self.upscale_diffusion_negative_prompt
                            ),
                            "strength": get_item_for_scale(self.upscale_diffusion_strength),
                            "num_inference_steps": get_item_for_scale(self.upscale_diffusion_steps),
                            "guidance_scale": get_item_for_scale(
                                self.upscale_diffusion_guidance_scale
                            ),
                            "chunking_size": self.upscale_diffusion_chunking_size,
                            "chunking_blur": self.upscale_diffusion_chunking_blur,
                            "progress_callback": progress_callback,
                        }
                        if self.upscale_diffusion_scale_chunking_size:
                            # Max out at half of the frame size or we get discontinuities
                            kwargs["chunking_size"] = min(
                                self.upscale_diffusion_chunking_size * (j + 1), pipeline.size // 2
                            )
                        if self.upscale_diffusion_scale_chunking_blur:
                            kwargs["chunking_blur"] = min(
                                self.upscale_diffusion_chunking_blur * (j + 1), pipeline.size // 2
                            )
                        upscale_controlnet = get_item_for_scale(self.upscale_diffusion_controlnet)
                        if upscale_controlnet is not None:
                            logger.debug(f"Enabling {upscale_controlnet} for upscale diffusion")
                            if upscale_controlnet == "canny":
                                pipeline.controlnet = "canny"
                                kwargs["control_image"] = pipeline.edge_detector.canny(image)
                            elif upscale_controlnet == "mlsd":
                                pipeline.controlnet = "mlsd"
                                kwargs["control_image"] = pipeline.edge_detector.mlsd(image)
                            elif upscale_controlnet == "hed":
                                pipeline.controlnet = "hed"
                                kwargs["control_image"] = pipeline.edge_detector.hed(image)
                            elif upscale_controlnet == "tile":
                                pipeline.controlnet = "tile"
                                kwargs["control_image"] = image
                            else:
                                logger.error(f"Unknown controlnet {upscale_controlnet}, ignoring.")
                                pipeline.controlnet = None
                        else:
                            pipeline.controlnet = None
                        image = pipeline(**kwargs).images[0]
                        images[i] = image
                    if re_enable_safety:
                        pipeline.safe = True
                if j < len(scales) - 1 and image_callback is not None:
                    image_callback(images)
        return StableDiffusionPipelineOutput(images=images, nsfw_content_detected=nsfw)

    def prepare_pipeline(self, pipeline: DiffusionPipelineManager) -> None:
        """
        Assigns pipeline-level variables.
        """
        pipeline.model = self.model
        pipeline.lora = self.lora
        pipeline.inversion = self.inversion
        pipeline.size = self.size
        if self.build_tensorrt:
            pipeline.build_tensorrt = True

    def execute_nodes(
        self,
        pipeline: DiffusionPipelineManager,
        progress_callback: Optional[Callable[[int, int, float], None]] = None,
        image_callback: Optional[Callable[[List[PIL.Image.Image]], None]] = None,
        image_callback_steps: Optional[int] = None,
    ) -> Tuple[List[PIL.Image.Image], List[bool]]:
        """
        This is called during execute(). It will go through the steps in order and return
        the resulting image(s).
        """
        if not self.nodes:
            if not self.image:
                raise ValueError("No image and no steps; cannot execute plan.")
            return [self.image], [False]

        # Define callback kwargs
        callback_kwargs = {
            "progress_callback": progress_callback,
            "latent_callback_steps": image_callback_steps,
            "latent_callback_type": "pil",
        }

        # Set up the pipeline
        self.prepare_pipeline(pipeline)

        if self.seed is not None:
            # Set up the RNG
            pipeline.seed = self.seed

        images = [PIL.Image.new("RGBA", (self.width, self.height)) for i in range(self.samples)]
        image_draw = [PIL.ImageDraw.Draw(image) for image in images]
        nsfw_content_detected = [False] * self.samples

        # Keep a final mask of all nodes to outpaint in the end
        outpaint_mask = PIL.Image.new("RGB", (self.width, self.height), (255, 255, 255))
        outpaint_draw = PIL.ImageDraw.Draw(outpaint_mask)

        for i, node in enumerate(self.nodes):
            if image_callback is not None:

                def node_image_callback(callback_images: List[PIL.Image.Image]) -> None:
                    for j, callback_image in enumerate(callback_images):
                        images[j].paste(node.resize_image(callback_image), node.bounds[0])
                    image_callback(images)  # type: ignore

            else:
                node_image_callback = None  # type: ignore
            invocation_kwargs = {**self.kwargs, **callback_kwargs}
            result = node.execute(
                pipeline, latent_callback=node_image_callback, **invocation_kwargs
            )
            for j, image in enumerate(result["images"]):
                image = node.resize_image(image)
                if image.mode == "RGBA":
                    # Draw the alpha mask of the return image onto the outpaint mask
                    alpha = image.split()[-1]
                    black = PIL.Image.new("RGB", alpha.size, (0, 0, 0))
                    outpaint_mask.paste(black, node.bounds[0], mask=alpha)
                    image_draw[j].rectangle((*node.bounds[0], *node.bounds[1]), fill=(0, 0, 0, 0))
                    images[j].paste(image, node.bounds[0], mask=alpha)
                else:
                    # Draw a rectangle directly
                    outpaint_draw.rectangle(node.bounds, fill="#000000")
                    images[j].paste(node.resize_image(image), node.bounds[0])

                nsfw_content_detected[j] = nsfw_content_detected[j] or (
                    "nsfw_content_detected" in result and result["nsfw_content_detected"][j]
                )

            # Call the callback
            if image_callback is not None:
                image_callback(images)

        # Determine if there's anything left to outpaint
        image_r_min, image_r_max = outpaint_mask.getextrema()[1]
        if image_r_max > 0:
            # Outpaint
            del invocation_kwargs["num_images_per_prompt"]
            outpaint_mask = feather_mask(outpaint_mask)

            outpaint_prompt_tokens = TokenMerger()
            outpaint_negative_prompt_tokens = TokenMerger()

            for i, node in enumerate(self.nodes):
                if node.step.prompt is not None:
                    outpaint_prompt_tokens.add(node.step.prompt)
                if node.step.negative_prompt is not None:
                    outpaint_negative_prompt_tokens.add(node.step.negative_prompt)

            if self.prompt is not None:
                outpaint_prompt_tokens.add(self.prompt, 2)  # Weighted
            if self.negative_prompt is not None:
                outpaint_negative_prompt_tokens.add(self.negative_prompt, 2)

            for i, image in enumerate(images):
                pipeline.controlnet = None
                result = pipeline(
                    image=image,
                    mask=outpaint_mask,
                    prompt=str(outpaint_prompt_tokens),
                    negative_prompt=str(outpaint_negative_prompt_tokens),
                    latent_callback=image_callback,
                    num_images_per_prompt=1,
                    **invocation_kwargs,
                )
                images[i] = result["images"][0]
                nsfw_content_detected[i] = nsfw_content_detected[i] or (
                    "nsfw_content_detected" in result and result["nsfw_content_detected"][0]
                )

        return images, nsfw_content_detected

    def get_serialization_dict(self) -> Dict[str, Any]:
        """
        Serializes the whole plan for storage or passing between processes.
        """
        upscale_dict: Optional[Dict[str, Any]] = None
        if self.upscale:
            upscale_diffusion_dict: Union[bool, Dict[str, Any]] = self.upscale_diffusion
            if self.upscale_diffusion:
                upscale_diffusion_dict = {
                    "steps": self.upscale_diffusion_steps,
                    "guidance_scale": self.upscale_diffusion_guidance_scale,
                    "chunking_size": self.upscale_diffusion_chunking_size,
                    "chunking_blur": self.upscale_diffusion_chunking_blur,
                    "strength": self.upscale_diffusion_strength,
                    "prompt": self.upscale_diffusion_prompt,
                    "negative_prompt": self.upscale_diffusion_negative_prompt,
                    "controlnet": self.upscale_diffusion_controlnet,
                    "scale_chunking_size": self.upscale_diffusion_scale_chunking_size,
                    "scale_chunking_blur": self.upscale_diffusion_scale_chunking_blur,
                }
            upscale_dict = {
                "method": self.upscale,
                "amount": self.outscale,
                "iterative": self.upscale_iterative,
                "diffusion": upscale_diffusion_dict,
            }

        return {
            "model": self.model,
            "lora": self.lora,
            "inversion": self.inversion,
            "width": self.width,
            "height": self.height,
            "size": self.size,
            "seed": self.seed,
            "prompt": self.prompt,
            "negative_prompt": self.negative_prompt,
            "image": self.image,
            "image_callback_steps": self.image_callback_steps,
            "nodes": [node.get_serialization_dict() for node in self.nodes],
            "samples": self.samples,
            "upscale": upscale_dict,
            "chunking_size": self.chunking_size,
            "chunking_blur": self.chunking_blur,
            "build_tensorrt": self.build_tensorrt,
        }

    @staticmethod
    def deserialize_dict(plan_dict: Dict[str, Any]) -> DiffusionPlan:
        """
        Given a serialized dictionary, instantiate a diffusion plan.
        """
        kwargs = {
            "model": plan_dict["model"],
            "nodes": [
                DiffusionNode.deserialize_dict(node_dict)
                for node_dict in plan_dict.get("nodes", [])
            ],
        }

        for arg in [
            "size",
            "lora",
            "inversion",
            "width",
            "height",
            "image_callback_steps",
            "image",
            "chunking_size",
            "chunking_blur",
            "samples",
            "seed",
            "prompt",
            "negative_prompt",
            "build_tensorrt",
        ]:
            if arg in plan_dict:
                kwargs[arg] = plan_dict[arg]

        upscale = plan_dict.get("upscale", None)
        if isinstance(upscale, dict):
            kwargs["upscale"] = upscale["method"]
            kwargs["outscale"] = upscale["amount"]
            kwargs["upscale_iterative"] = upscale.get("iterative", False)
            upscale_diffusion = upscale.get("diffusion", None)
            if isinstance(upscale_diffusion, dict):
                kwargs["upscale_diffusion"] = True
                for arg in [
                    "steps",
                    "controlnet",
                    "guidance_scale",
                    "strength",
                    "chunking_size",
                    "chunking_blur",
                    "prompt",
                    "negative_prompt",
                    "scale_chunking_size",
                    "scale_chunking_blur",
                ]:
                    if arg in upscale_diffusion:
                        kwargs[f"upscale_diffusion_{arg}"] = upscale_diffusion[arg]
        result = DiffusionPlan(**kwargs)
        return result

    @staticmethod
    def create_mask(
        width: int, height: int, left: int, top: int, right: int, bottom: int
    ) -> PIL.Image.Image:
        """
        Creates a mask from 6 dimensions
        """
        image = PIL.Image.new("RGB", (width, height))
        draw = PIL.ImageDraw.Draw(image)
        draw.rectangle([(left, top), (right, bottom)], fill="#ffffff")
        return image

    @staticmethod
    def from_nodes(
        size: Optional[int] = DEFAULT_SIZE,
        model: Optional[str] = None,
        lora: Optional[
            Union[str, List[str], Tuple[str, float], List[Union[str, Tuple[str, float]]]]
        ] = None,
        inversion: Optional[Union[str, List[str]]] = None,
        model_prompt: Optional[str] = None,
        model_negative_prompt: Optional[str] = None,
        samples: int = 1,
        seed: Optional[int] = None,
        width: Optional[int] = DEFAULT_SIZE,
        height: Optional[int] = DEFAULT_SIZE,
        image_callback_steps: int = DEFAULT_IMAGE_CALLBACK_STEPS,
        nodes: List[NodeDict] = [],
        chunking_size: Optional[int] = DEFAULT_CHUNKING_SIZE,
        chunking_blur: Optional[int] = DEFAULT_CHUNKING_BLUR,
        prompt: Optional[str] = None,
        negative_prompt: Optional[str] = None,
        num_inference_steps: Optional[int] = DEFAULT_INFERENCE_STEPS,
        guidance_scale: Optional[float] = DEFAULT_GUIDANCE_SCALE,
        outscale: Optional[int] = 1,
        upscale: Optional[
            Union[
                Literal[
                    "esrgan",
                    "esrganime",
                    "gfpgan",
                    "lanczos",
                    "bilinear",
                    "bicubic",
                    "lanczos",
                    "nearest",
                ],
                List[
                    Literal[
                        "esrgan",
                        "esrganime",
                        "gfpgan",
                        "lanczos",
                        "bilinear",
                        "bicubic",
                        "lanczos",
                        "nearest",
                    ]
                ],
            ]
        ] = None,
        upscale_diffusion: bool = False,
        upscale_iterative: bool = False,
        upscale_diffusion_steps: Optional[Union[int, List[int]]] = DEFAULT_UPSCALE_DIFFUSION_STEPS,
        upscale_diffusion_guidance_scale: Optional[
            Union[float, int, List[Union[float, int]]]
        ] = DEFAULT_UPSCALE_DIFFUSION_GUIDANCE_SCALE,
        upscale_diffusion_strength: Optional[
            Union[float, List[float]]
        ] = DEFAULT_UPSCALE_DIFFUSION_STRENGTH,
        upscale_diffusion_prompt: Optional[Union[str, List[str]]] = DEFAULT_UPSCALE_PROMPT,
        upscale_diffusion_negative_prompt: Optional[
            Union[str, List[str]]
        ] = DEFAULT_UPSCALE_NEGATIVE_PROMPT,
        upscale_diffusion_controlnet: Optional[
            Union[
                Literal["canny", "tile", "mlsd", "hed", "scribble", "inpaint"],
                List[Literal["canny", "tile", "mlsd", "hed", "scribble", "inpaint"]],
            ]
        ] = None,
        upscale_diffusion_chunking_size: Optional[int] = DEFAULT_UPSCALE_DIFFUSION_CHUNKING_SIZE,
        upscale_diffusion_chunking_blur: Optional[int] = DEFAULT_UPSCALE_DIFFUSION_CHUNKING_BLUR,
        upscale_diffusion_scale_chunking_size: bool = True,
        upscale_diffusion_scale_chunking_blur: bool = True,
    ) -> DiffusionPlan:
        """
        Assembles a diffusion plan from step dictionaries.
        """
        # First instantiate the plan
        plan = DiffusionPlan(
            model=model,
            lora=lora,
            inversion=inversion,
            samples=samples,
            size=size,
            seed=seed,
            width=width,
            height=height,
            image_callback_steps=image_callback_steps,
            upscale=upscale,
            outscale=outscale,
            prompt=prompt,
            chunking_size=chunking_size,
            chunking_blur=chunking_blur,
            negative_prompt=negative_prompt,
            upscale_iterative=upscale_iterative,
            upscale_diffusion=upscale_diffusion,
            upscale_diffusion_steps=upscale_diffusion_steps,
            upscale_diffusion_guidance_scale=upscale_diffusion_guidance_scale,
            upscale_diffusion_strength=upscale_diffusion_strength,
            upscale_diffusion_controlnet=upscale_diffusion_controlnet,
            upscale_diffusion_chunking_size=upscale_diffusion_chunking_size,
            upscale_diffusion_chunking_blur=upscale_diffusion_chunking_blur,
            upscale_diffusion_scale_chunking_size=upscale_diffusion_scale_chunking_size,
            upscale_diffusion_scale_chunking_blur=upscale_diffusion_scale_chunking_blur,
            nodes=[],
        )

        # We'll assemble multiple token sets for overall diffusion
        upscale_diffusion_prompt_tokens = [TokenMerger()]
        upscale_diffusion_negative_prompt_tokens = [TokenMerger()]

        if upscale_diffusion_prompt:
            if isinstance(upscale_diffusion_prompt, list):
                upscale_diffusion_prompt = [
                    upscale_prompt for upscale_prompt in upscale_diffusion_prompt if upscale_prompt
                ]
                for i, upscale_prompt in enumerate(upscale_diffusion_prompt):
                    if len(upscale_diffusion_prompt_tokens) < i + 1:
                        upscale_diffusion_prompt_tokens.append(TokenMerger())
                    upscale_diffusion_prompt_tokens[i].add(upscale_prompt)
            else:
                upscale_diffusion_prompt_tokens[0].add(upscale_diffusion_prompt)
        if prompt:
            for token_merger in upscale_diffusion_prompt_tokens:
                token_merger.add(prompt, GLOBAL_PROMPT_UPSCALE_WEIGHT)
        if model_prompt:
            for token_merger in upscale_diffusion_prompt_tokens:
                token_merger.add(model_prompt, MODEL_PROMPT_WEIGHT)

        if upscale_diffusion_negative_prompt:
            if isinstance(upscale_diffusion_negative_prompt, list):
                upscale_diffusion_negative_prompt = [
                    upscale_negative_prompt
                    for upscale_negative_prompt in upscale_diffusion_negative_prompt
                    if upscale_negative_prompt
                ]
                for i, negative_prompt in enumerate(upscale_diffusion_negative_prompt):
                    if len(upscale_diffusion_negative_prompt_tokens) < i + 1:
                        upscale_diffusion_negative_prompt_tokens.append(TokenMerger())
                    upscale_diffusion_negative_prompt_tokens[i].add(negative_prompt)
            else:
                upscale_diffusion_negative_prompt_tokens[0].add(upscale_diffusion_negative_prompt)
        if negative_prompt:
            for token_merger in upscale_diffusion_negative_prompt_tokens:
                token_merger.add(negative_prompt, GLOBAL_PROMPT_UPSCALE_WEIGHT)
        if model_negative_prompt:
            for token_merger in upscale_diffusion_negative_prompt_tokens:
                token_merger.add(model_negative_prompt, MODEL_PROMPT_WEIGHT)

        # Now assemble the diffusion steps
        node_count = len(nodes)
        if node_count == 0:
            # Basic txt2img
            prompt_tokens = TokenMerger()
            if prompt:
                prompt_tokens.add(prompt)
            if model_prompt:
                prompt_tokens.add(model_prompt, MODEL_PROMPT_WEIGHT)

            negative_prompt_tokens = TokenMerger()
            if negative_prompt:
                negative_prompt_tokens.add(negative_prompt)
            if model_negative_prompt:
                negative_prompt_tokens.add(model_negative_prompt, MODEL_PROMPT_WEIGHT)

            step = DiffusionStep(
                width=width,
                height=height,
                prompt=str(prompt_tokens),
                negative_prompt=str(negative_prompt_tokens),
                num_inference_steps=num_inference_steps,
                guidance_scale=guidance_scale,
            )

            plan.nodes = [DiffusionNode([(0, 0), (plan.width, plan.height)], step)]

            plan.upscale_diffusion_prompt = [
                str(merger) for merger in upscale_diffusion_prompt_tokens
            ]
            plan.upscale_diffusion_negative_prompt = [
                str(merger) for merger in upscale_diffusion_negative_prompt_tokens
            ]
            return plan

        # Using the diffusion canvas, assemble a multi-step plan
        for i, node_dict in enumerate(nodes):
            step = DiffusionStep(
                num_inference_steps=num_inference_steps, guidance_scale=guidance_scale
            )

            node_width = int(node_dict["w"])
            node_height = int(node_dict["h"])
            node_left = int(node_dict["x"])
            node_top = int(node_dict["y"])
            node_fit = node_dict.get("fit", None)
            node_anchor = node_dict.get("anchor", None)
            node_bounds = [
                (node_left, node_top),
                (node_left + node_width, node_top + node_height),
            ]

            node_infer = node_dict.get("infer", False)
            node_inpaint = node_dict.get("inpaint", False)
            node_control = node_dict.get("control", False)

            node_controlnet = node_dict.get("controlnet", None)
            node_prompt = node_dict.get("prompt", None)
            node_negative_prompt = node_dict.get("negative_prompt", None)
            node_strength: Optional[float] = node_dict.get("strength", None)
            node_conditioning_scale: Optional[float] = node_dict.get("conditioning_scale", None)
            node_image = node_dict.get("image", None)
            node_inpaint_mask = node_dict.get("mask", None)
            node_process_control_image = node_dict.get("process_control_image", True)
            node_scale_to_model_size = node_dict.get("scale_to_model_size", True)
            node_remove_background = bool(node_dict.get("remove_background", False))
            node_inference_steps: Optional[int] = node_dict.get("inference_steps", None)  # type: ignore[assignment]
            node_guidance_scale: Optional[float] = node_dict.get("guidance_scale", None)  # type: ignore[assignment]

            node_prompt_tokens = TokenMerger()
            node_negative_prompt_tokens = TokenMerger()

            if node_prompt:
                node_prompt_tokens.add(node_prompt)
                for merger in upscale_diffusion_prompt_tokens:
                    merger.add(node_prompt, UPSCALE_PROMPT_STEP_WEIGHT / node_count)
            if prompt and node_image:
                # Only add global prompt to image nodes, it overrides too much on region nodes
                node_prompt_tokens.add(prompt, GLOBAL_PROMPT_STEP_WEIGHT)
            if model_prompt:
                node_prompt_tokens.add(model_prompt, MODEL_PROMPT_WEIGHT)

            if node_negative_prompt:
                node_negative_prompt_tokens.add(node_negative_prompt)
                for merger in upscale_diffusion_negative_prompt_tokens:
                    merger.add(node_negative_prompt, UPSCALE_PROMPT_STEP_WEIGHT / node_count)
            if negative_prompt and node_image:
                # Only add global prompt to image nodes, it overrides too much on region nodes
                node_negative_prompt_tokens.add(negative_prompt, GLOBAL_PROMPT_STEP_WEIGHT)
            if model_negative_prompt:
                node_negative_prompt_tokens.add(model_negative_prompt, MODEL_PROMPT_WEIGHT)

            if node_image:
                if node_remove_background:
                    node_image = remove_background(node_image)

                # Resize node image to nearest multiple of 8
                node_image_width, node_image_height = node_image.size
                node_image_width = 8 * math.ceil(node_image_width / 8)
                node_image_height = 8 * math.ceil(node_image_height / 8)
                node_image = node_image.resize((node_image_width, node_image_height))

                node_image = fit_image(node_image, node_width, node_height, node_fit, node_anchor)
                node_mask = PIL.Image.new("RGB", node_image.size, (255, 255, 255))
                node_alpha = node_image.split()[-1]
                node_alpha_clamp = PIL.Image.eval(node_alpha, lambda a: 255 if a > 128 else 0)

                black = PIL.Image.new("RGB", node_image.size, (0, 0, 0))
                white = PIL.Image.new("RGB", node_image.size, (255, 255, 255))
                node_mask.paste(black, mask=node_alpha)
                node_mask_r_min, node_mask_r_max = node_mask.getextrema()[1]

                node_image_needs_outpainting = node_mask_r_max > 0

                if node_inpaint and node_inpaint_mask:
                    # Inpaint prior to anything else. First invert mask
                    node_inpaint_mask = PIL.ImageOps.invert(node_inpaint_mask.convert("L"))
                    if node_image_needs_outpainting:
                        # Merge inpaint and outpaint masks
                        node_mask.paste(node_inpaint_mask)
                        node_inverse_alpha = PIL.Image.eval(node_alpha_clamp, lambda a: 255 - a)
                        node_mask.paste(white, mask=node_inverse_alpha)
                    else:
                        # Just use inpaint mask
                        node_mask.paste(node_inpaint_mask)

                    inpaint_image_step = DiffusionStep(
                        image=node_image,
                        mask=feather_mask(node_mask.convert("1")),
                        prompt=str(node_prompt_tokens),
                        negative_prompt=str(node_negative_prompt_tokens),
                        guidance_scale=guidance_scale,
                    )
                    step_image = inpaint_image_step
                elif node_image_needs_outpainting and (node_infer or node_control):
                    # There are gaps; add an outpaint step before the infer/control
                    outpaint_image_step = DiffusionStep(
                        image=node_image,
                        mask=feather_mask(node_mask.convert("1")),
                        prompt=str(node_prompt_tokens),
                        negative_prompt=str(node_negative_prompt_tokens),
                        guidance_scale=guidance_scale,
                    )
                    step_image = outpaint_image_step
                else:
                    step_image = node_image

                if not node_infer and not node_inpaint and not node_control:
                    # Just paste image
                    step.image = step_image
                elif not node_infer and node_inpaint and not node_control:
                    # Just inpaint image
                    step = step_image
                else:
                    # Some combination of ops
                    step.prompt = str(node_prompt_tokens)
                    step.negative_prompt = str(node_negative_prompt_tokens)
                    if node_strength:
                        step.strength = node_strength
                    if node_infer:
                        step.image = step_image
                    if node_control and node_controlnet:
                        step.controlnet = node_controlnet  # type: ignore
                        step.control_image = step_image
                        if not node_process_control_image:
                            step.process_control_image = False
                        if node_conditioning_scale:
                            step.conditioning_scale = node_conditioning_scale
            elif node_prompt:
                step.prompt = str(node_prompt_tokens)
                step.negative_prompt = str(node_negative_prompt_tokens)
                step.width = node_width
                step.height = node_height
                step.remove_background = node_remove_background
            else:
                raise ValueError("Can't assemble a node from arguments")

            # Set common args for node
            if node_inference_steps:
                step.num_inference_steps = node_inference_steps
            if node_guidance_scale:
                step.guidance_scale = node_guidance_scale
            if not node_scale_to_model_size:
                step.scale_to_model_size = False

            # Add step to plan
            plan.nodes.append(DiffusionNode(node_bounds, step))

        plan.upscale_diffusion_prompt = [str(prompt) for prompt in upscale_diffusion_prompt_tokens]
        plan.upscale_diffusion_negative_prompt = [
            str(negative_prompt) for negative_prompt in upscale_diffusion_negative_prompt_tokens
        ]
        return plan



# type: ignore
# Adapted from https://github.com/ashukid/hed-edge-detector/blob/master/edge.py

from typing import Any, Sequence, Tuple, List

__all__ = ["HEDCropLayer"]


class HEDCropLayer:
    """
    Added to the CV2 NN during HED edge detection.
    """

    def __init__(self, *args: Any) -> None:
        self.x_start, self.x_end = 0, 0
        self.y_start, self.y_end = 0, 0

    def getMemoryShapes(self, inputs: Sequence[Tuple[int, ...]]) -> List[List[int]]:
        """
        Gets the shapes from the input layer.
        We received two inputs. We need to crop the first blob to match the shape of the second.
        """
        input_shape, target_shape = inputs[0:2]
        batch_size, num_channels, input_height, input_width = input_shape[0:4]
        target_height, target_width = target_shape[2:4]

        self.y_start = (input_height - target_height) // 2
        self.x_start = (input_width - target_width) // 2
        self.y_end = self.y_start + target_height
        self.x_end = self.x_start + target_width

        return [[batch_size, num_channels, target_height, target_width]]

    def forward(self, inputs: Sequence[Sequence[Sequence]]) -> List[Sequence]:
        """
        On forward only send first input blob.
        """
        return [inputs[0][:, :, self.y_start : self.y_end, self.x_start : self.x_end]]



import os
from typing import List, Dict, Any

from webob import Request, Response

from pibble.api.exceptions import BadRequestError, StateConflictError
from pibble.ext.user.server.base import UserExtensionHandlerRegistry

from enfugue.util import check_make_directory
from enfugue.partner.civitai import CivitAI
from enfugue.api.controller.base import EnfugueAPIControllerBase

__all__ = ["EnfugueAPIDownloadsController"]


class EnfugueAPIDownloadsController(EnfugueAPIControllerBase):
    handlers = UserExtensionHandlerRegistry()

    @property
    def civitai(self) -> CivitAI:
        """
        Gets the client. Instantiates if necessary.
        """
        if not hasattr(self, "_civitai"):
            self._civitai = CivitAI()
            self._civitai.configure()
        return self._civitai

    @handlers.path("^/api/download$")
    @handlers.methods("POST")
    @handlers.format()
    @handlers.secured("Download", "create")
    def start_download(self, request: Request, response: Response) -> Dict[str, Any]:
        """
        Starts a download.
        """
        try:
            target_dir = os.path.join(self.engine_root, str(request.parsed["type"]))
            target_file = os.path.join(target_dir, request.parsed["filename"])

            if os.path.exists(target_file) and not request.parsed.get("overwrite", False):
                raise StateConflictError(f"File exists: {request.parsed['filename']}")
            
            check_make_directory(target_dir)
            
            return self.manager.download(
                request.token.user.id, request.parsed["url"], target_file
            ).format()
        except KeyError as ex:
            raise BadRequestError(f"Missing required argument {ex}")

    @handlers.path("^/api/download$")
    @handlers.methods("GET")
    @handlers.secured()
    @handlers.format()
    def get_download_status(self, request: Request, response: Response) -> List[Dict[str, Any]]:
        """
        Gets status for downloads. Tese are only stored in memory.
        """
        return [download.format() for download in self.manager.get_downloads(request.token.user.id)]

    @handlers.path("^/api/civitai/(?P<lookup>[a-z]+)$")
    @handlers.methods("GET")
    @handlers.secured()
    @handlers.format()
    def civitai_lookup(
        self, request: Request, response: Response, lookup: str
    ) -> List[Dict[str, Any]]:
        """
        Performs a lookup in CivitAI
        """
        lookup_type = {
            "checkpoint": "Checkpoint",
            "inversion": "TextualInversion",
            "lora": "LORA",
            "controlnet": "Controlnet",
            "poses": "Poses",
            "hypernetwork": "Hypetnetwork",
            "gradient": "AestheticGradient",
        }.get(lookup, None)

        if lookup_type is None:
            raise BadRequestError(f"Unknown lookup type {lookup_type}")

        query = request.params.get("query", None)

        lookup_kwargs = {
            "types": lookup_type,
            "limit": 20,
            "nsfw": not self.safe,
            "query": request.params.get("query", None),
            "page": request.params.get("page", None),
            "sort": request.params.get("sort", "Most Downloaded"),
            "period": request.params.get("period", "Month"),
            "allowCommercialUse": request.params.get("allow_commercial_use", None),
        }

        iterator = self.civitai.get_models(**lookup_kwargs)  # type: ignore
        return [result.serialize() for result in iterator]



from enfugue.diffusion.rt.engine import Engine
from enfugue.diffusion.rt.optimizer import Optimizer
from enfugue.diffusion.rt.pipeline import EnfugueTensorRTStableDiffusionPipeline

Engine, Optimizer, EnfugueTensorRTStableDiffusionPipeline

__all__ = ["Engine", "Optimizer", "EnfugueTensorRTStableDiffusionPipeline"]



import torch
import tensorrt as trt

from typing import Optional, List, Dict, Iterator, Any, Union, Tuple
from typing_extensions import Self

from contextlib import contextmanager

from polygraphy import cuda
from transformers import CLIPImageProcessor, CLIPTextModel, CLIPTokenizer

from diffusers.schedulers import KarrasDiffusionSchedulers
from diffusers.models import AutoencoderKL, UNet2DConditionModel, ControlNetModel

from diffusers.pipelines.stable_diffusion import StableDiffusionSafetyChecker

from enfugue.util import logger
from enfugue.diffusion.pipeline import EnfugueStableDiffusionPipeline
from enfugue.diffusion.util import DTypeConverter
from enfugue.diffusion.rt.engine import Engine
from enfugue.diffusion.rt.model import BaseModel, UNet, VAE, CLIP, ControlledUNet, ControlNet


class EnfugueTensorRTStableDiffusionPipeline(EnfugueStableDiffusionPipeline):
    models: Dict[str, BaseModel]
    engine: Dict[str, Engine]
    unet: UNet2DConditionModel

    def __init__(
        self,
        vae: AutoencoderKL,
        text_encoder: CLIPTextModel,
        tokenizer: CLIPTokenizer,
        unet: UNet2DConditionModel,
        controlnet: Optional[ControlNetModel],
        scheduler: KarrasDiffusionSchedulers,
        safety_checker: StableDiffusionSafetyChecker,
        feature_extractor: CLIPImageProcessor,
        requires_safety_checker: bool = True,
        engine_size: int = 512,  # Recommended even for machines that can handle more
        chunking_size: int = 32,
        chunking_blur: int = 64,
        max_batch_size: int = 16,
        # ONNX export parameters
        force_engine_rebuild: bool = False,
        vae_engine_dir: Optional[str] = None,
        clip_engine_dir: Optional[str] = None,
        unet_engine_dir: Optional[str] = None,
        controlled_unet_engine_dir: Optional[str] = None,
        controlnet_engine_dir: Optional[str] = None,
        build_static_batch: bool = False,
        build_dynamic_shape: bool = False,
        build_preview_features: bool = False,
        onnx_opset: int = 17,
    ) -> None:
        super(EnfugueTensorRTStableDiffusionPipeline, self).__init__(
            vae,
            text_encoder,
            tokenizer,
            unet,
            controlnet,
            scheduler,
            safety_checker,
            feature_extractor,
            requires_safety_checker,
            engine_size,
            chunking_size,
            chunking_blur,
        )

        if self.controlnet is not None:
            # Hijack forward
            self.unet.forward = self.controlled_unet_forward
        self.vae.forward = self.vae.decode
        self.onnx_opset = onnx_opset
        self.force_engine_rebuild = force_engine_rebuild
        self.vae_engine_dir = vae_engine_dir
        self.clip_engine_dir = clip_engine_dir
        self.unet_engine_dir = unet_engine_dir
        self.controlled_unet_engine_dir = controlled_unet_engine_dir
        self.controlnet_engine_dir = controlnet_engine_dir
        self.build_static_batch = build_static_batch
        self.build_dynamic_shape = build_dynamic_shape
        self.build_preview_features = build_preview_features
        self.max_batch_size = max_batch_size

        # TODO: Restrict batch size to 4 for larger image dimensions as a WAR for TensorRT limitation.
        if self.build_dynamic_shape or self.engine_size > 512:
            self.max_batch_size = 4

        self.stream = None  # loaded in load_resources()
        self.models = {}  # loaded in load_models()
        self.engine = {}  # loaded in build_engines()

    @staticmethod
    def device_view(t: torch.Tensor) -> cuda.DeviceView:
        """
        Gets a device view over a tensor
        """
        return cuda.DeviceView(
            ptr=t.data_ptr(), shape=t.shape, dtype=DTypeConverter.from_torch(t.dtype)
        )

    def controlled_unet_forward(self, *args, **kwargs):
        """
        Highjacks unet forwarding.
        """
        sample, timestep, embeddings, mid_block = args[:4]
        down_blocks = args[4:]
        return UNet2DConditionModel.forward(
            self.unet,
            sample,
            timestep,
            embeddings,
            mid_block_additional_residual=mid_block,
            down_block_additional_residuals=down_blocks,
        )

    def load_models(self, use_fp16: bool = False) -> None:
        """
        Loads pipeline models to build later.
        """
        self.embedding_dim = self.text_encoder.config.hidden_size
        models_args = {
            "device": self.torch_device,
            "max_batch_size": self.max_batch_size,
            "embedding_dim": self.embedding_dim,
            "use_fp16": use_fp16,
        }
        if self.clip_engine_dir is not None:
            self.models["clip"] = CLIP(self.text_encoder, **models_args)
        if self.unet_engine_dir is not None:
            self.models["unet"] = UNet(
                self.unet, unet_dim=self.unet.config.in_channels, **models_args
            )
        if self.controlled_unet_engine_dir is not None:
            self.models["controlledunet"] = ControlledUNet(
                self.unet, unet_dim=self.unet.config.in_channels, **models_args
            )
        if self.controlnet_engine_dir is not None and self.controlnet is not None:
            self.models["controlnet"] = ControlNet(
                self.controlnet, unet_dim=self.unet.config.in_channels, **models_args
            )
        if self.vae_engine_dir is not None:
            self.models["vae"] = VAE(self.vae, **models_args)

    def load_resources(self, image_height: int, image_width: int, batch_size: int) -> None:
        """
        Allocates buffers for configured RT models.
        """
        self.stream = cuda.Stream()

        # Allocate buffers for TensorRT engine bindings
        for model_name, obj in self.models.items():
            self.engine[model_name].allocate_buffers(
                shape_dict=obj.get_shape_dict(batch_size, image_height, image_width),
                device=self.torch_device,
            )

    @contextmanager
    def get_runtime_context(
        self, batch_size: int, device: Union[str, torch.device]
    ) -> Iterator[None]:
        """
        We initialize the TensorRT runtime here.
        """
        self.load_resources(self.engine_size, self.engine_size, batch_size)
        with (
            torch.inference_mode(),
            torch.autocast(device.type if isinstance(device, torch.device) else device),
            trt.Runtime(trt.Logger(trt.Logger.ERROR)),
        ):
            if self.unet_engine_dir is not None or self.controlled_unet_engine_dir is not None:
                logger.info("TensorRT unloading UNET")
                self.unet = self.unet.to("cpu")
            yield
            if self.unet_engine_dir is not None or self.controlled_unet_engine_dir is not None:
                logger.info("TensorRT reloading UNET")
                self.unet = self.unet.to(device)

    def to(
        self,
        torch_device: Optional[Union[str, torch.device]] = None,
        silence_dtype_warnings: bool = False,
        torch_dtype: torch.dtype = torch.float16,
    ) -> Self:
        """
        First follow the parent to(), then load engines.
        """
        super().to(torch_device, silence_dtype_warnings=silence_dtype_warnings)

        # set device
        self.torch_device = self._execution_device

        # load models
        use_fp16 = torch_dtype is torch.float16
        self.load_models(use_fp16)

        # build engines
        engines_to_build: Dict[str, BaseModel] = {}
        if self.clip_engine_dir is not None:
            engines_to_build[self.clip_engine_dir] = self.models["clip"]
        if self.unet_engine_dir is not None:
            engines_to_build[self.unet_engine_dir] = self.models["unet"]
        if self.controlled_unet_engine_dir is not None:
            engines_to_build[self.controlled_unet_engine_dir] = self.models["controlledunet"]
        if self.controlnet_engine_dir is not None:
            engines_to_build[self.controlnet_engine_dir] = self.models["controlnet"]
        if self.vae_engine_dir is not None:
            engines_to_build[self.vae_engine_dir] = self.models["vae"]

        self.engine = Engine.build_all(
            engines_to_build,
            self.onnx_opset,
            use_fp16=use_fp16,
            opt_image_height=self.engine_size,
            opt_image_width=self.engine_size,
            force_engine_rebuild=self.force_engine_rebuild,
            static_batch=self.build_static_batch,
            static_shape=not self.build_dynamic_shape,
            enable_preview=self.build_preview_features,
        )

        return self

    def create_latents(
        self,
        batch_size: int,
        num_channels_latents: int,
        height: int,
        width: int,
        dtype: Union[str, torch.dtype],
        device: Union[str, torch.device],
        generator: Optional[torch.Generator] = None,
    ) -> torch.Tensor:
        """
        Override to change to float32
        """
        return super(EnfugueTensorRTStableDiffusionPipeline, self).create_latents(
            batch_size,
            num_channels_latents,
            height,
            width,
            torch.float32,
            device,
            generator,
        )

    def encode_prompt(
        self,
        prompt: Optional[Union[str, List[str]]] = None,
        negative_prompt: Optional[Union[str, List[str]]] = None,
    ) -> torch.Tensor:
        """
        Encodes the prompt into text encoder hidden states.
        Args:
             prompt (`str` or `List[str]`, *optional*):
                prompt to be encoded
            negative_prompt (`str` or `List[str]`, *optional*):
                The prompt or prompts not to guide the image generation. If not defined, one has to pass
                `negative_prompt_embeds`. instead. If not defined, one has to pass `negative_prompt_embeds`. instead.
                Ignored when not using guidance (i.e., ignored if `guidance_scale` is less than `1`).
        """
        if "clip" not in self.engine:
            return super(EnfugueTensorRTStableDiffusionPipeline, self).encode_prompt(
                prompt, negative_prompt
            )

        # Tokenize prompt
        text_input_ids = (
            self.tokenizer(
                prompt,
                padding="max_length",
                max_length=self.tokenizer.model_max_length,
                truncation=True,
                return_tensors="pt",
            )
            .input_ids.type(torch.int32)
            .to(self.torch_device)
        )

        text_input_ids_inp = self.device_view(text_input_ids)
        # NOTE: output tensor for CLIP must be cloned because it will be overwritten when called again for negative prompt

        text_embeddings = (
            self.engine["clip"]
            .infer({"input_ids": text_input_ids_inp}, self.stream)["text_embeddings"]
            .clone()
        )

        # Tokenize negative prompt
        uncond_input_ids = (
            self.tokenizer(
                negative_prompt,
                padding="max_length",
                max_length=self.tokenizer.model_max_length,
                truncation=True,
                return_tensors="pt",
            )
            .input_ids.type(torch.int32)
            .to(self.torch_device)
        )

        uncond_input_ids_inp = self.device_view(uncond_input_ids)
        uncond_embeddings = self.engine["clip"].infer(
            {"input_ids": uncond_input_ids_inp}, self.stream
        )["text_embeddings"]

        # Concatenate the unconditional and text embeddings into a single batch to avoid doing two forward passes for classifier free guidance
        text_embeddings = torch.cat([uncond_embeddings, text_embeddings]).to(dtype=torch.float16)

        return text_embeddings

    def get_controlnet_conditioning_blocks(
        self,
        device: Union[str, torch.device],
        latents: torch.Tensor,
        timestep: torch.Tensor,
        encoder_hidden_states: torch.Tensor,
        controlnet_cond: Optional[torch.Tensor],
        conditioning_scale: float,
    ) -> Tuple[Optional[List[torch.Tensor]], Optional[torch.Tensor]]:
        """
        Executes the controlnet inference
        """
        if "controlnet" not in self.engine:
            return super(
                EnfugueTensorRTStableDiffusionPipeline, self
            ).get_controlnet_conditioning_blocks(
                device=device,
                latents=latents,
                timestep=timestep,
                encoder_hidden_states=encoder_hidden_states,
                controlnet_cond=controlnet_cond,
                conditioning_scale=conditioning_scale,
            )
        if controlnet_cond is None:
            return None, None

        timestep_float = timestep.float() if timestep.dtype != torch.float32 else timestep
        conditioning_scale_tensor = torch.Tensor([conditioning_scale]).to(
            dtype=torch.float32, device=device
        )
        inference_kwargs = {
            "sample": self.device_view(latents),
            "timestep": self.device_view(timestep_float),
            "encoder_hidden_states": self.device_view(encoder_hidden_states),
            "controlnet_cond": self.device_view(controlnet_cond),
            "conditioning_scale": self.device_view(conditioning_scale_tensor),
        }

        inferred = self.engine["controlnet"].infer(inference_kwargs, self.stream)
        return [inferred[f"down_block_{i}"] for i in range(12)], inferred["mid_block"]

    def predict_noise_residual(
        self,
        latents: torch.Tensor,
        timestep: torch.Tensor,
        embeddings: torch.Tensor,
        cross_attention_kwargs: Optional[Dict[str, Any]] = None,
        down_block_additional_residuals: Optional[List[torch.Tensor]] = None,
        mid_block_additional_residual: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Runs the RT engine inference on the unet
        """
        engine_name = "controlledunet"
        if engine_name not in self.engine:
            engine_name = "unet"
        if engine_name not in self.engine:
            return super(EnfugueTensorRTStableDiffusionPipeline, self).predict_noise_residual(
                latents,
                timestep,
                embeddings,
                cross_attention_kwargs,
                down_block_additional_residuals,
                mid_block_additional_residual,
            )

        timestep_float = timestep.float() if timestep.dtype != torch.float32 else timestep

        inference_kwargs = {
            "sample": self.device_view(latents),
            "timestep": self.device_view(timestep_float),
            "encoder_hidden_states": self.device_view(embeddings),
        }

        if (
            down_block_additional_residuals is not None
            and mid_block_additional_residual is not None
        ):
            if engine_name == "unet":
                if not getattr(self, "_informed_of_bad_controlled_unet", False):
                    logger.error(
                        "Incorrect use of UNet/ControlledUNet models. Controlnet input was detected, but only the non-controlled UNet is loaded. Controlnet input will be ignored going forward."
                    )
                    self._informed_of_bad_controlled_unet = True
            else:
                inference_kwargs["mid_block"] = self.device_view(mid_block_additional_residual)
                for i, block in enumerate(down_block_additional_residuals):
                    inference_kwargs[f"down_block_{i}"] = self.device_view(block)
        return self.engine[engine_name].infer(inference_kwargs, self.stream)["latent"]



from __future__ import annotations
import os
import time
import datetime

from typing import Optional, Dict, List, Any, Tuple, TypedDict
from threading import Thread, Event
from multiprocessing import Lock

from pibble.api.configuration import APIConfiguration
from pibble.api.exceptions import TooManyRequestsError, BadRequestError
from pibble.util.numeric import human_size

from enfugue.api.downloads import Download
from enfugue.api.invocations import Invocation
from enfugue.diffusion.engine import DiffusionEngine
from enfugue.diffusion.plan import DiffusionPlan
from enfugue.diffusion.constants import DEFAULT_MODEL, DEFAULT_INPAINTING_MODEL
from enfugue.util import logger, check_make_directory

__all__ = ["SystemManagerThread", "SystemManager"]


class SystemManagerThread(Thread):
    """
    This thread simply executes periodic tasks on the manager.
    """

    def __init__(self, manager: SystemManager) -> None:
        super(SystemManagerThread, self).__init__()
        self.manager = manager
        self.stop_event = Event()

    def stop(self) -> None:
        """
        Set the stop event.
        """
        self.stop_event.set()

    @property
    def stopped(self) -> bool:
        """
        Returns true if the stop event is set.
        """
        return self.stop_event.is_set()

    def run(self) -> None:
        """
        The main thread loop.
        """
        while not self.stopped:
            if not self.manager.running:
                logger.info("Manager exited, stopping manager thread.")
                return
            self.manager.do_periodic_tasks()
            time.sleep(0.5)


class SystemManagerDownloadStatusDict(TypedDict):
    active: int
    queued: int
    total: int


class SystemManagerInvocationStatusDict(TypedDict):
    active: bool
    queued: int
    total: int


class SystemManagerStatusDict(TypedDict):
    downloads: SystemManagerDownloadStatusDict
    invocations: SystemManagerInvocationStatusDict


class SystemManager:
    """
    This single class serves to perform all necessary asynchronous processes
    for the API, to reduce headache in wrangling many queues and processes.
    """

    DEFAULT_MAX_CONCURRENT_DOWNLOADS = 2
    DEFAULT_MAX_QUEUED_DOWNLOADS = 10
    DEFAULT_MAX_QUEUED_INVOCATIONS = 2

    downloads: Dict[int, List[Download]]
    invocations: Dict[int, List[Invocation]]
    download_queue: List[Download]
    invocation_queue: List[Invocation]
    active_invocation: Optional[Invocation]

    def __init__(self, configuration: APIConfiguration) -> None:
        self.lock = Lock()
        self.active_invocation = None
        self.configuration = configuration
        self.engine = DiffusionEngine(self.configuration)
        self.downloads = {}
        self.invocations = {}
        self.download_queue = []
        self.invocation_queue = []
        self.thread = SystemManagerThread(self)
        self.running = False

    def start_monitor(self) -> None:
        """
        Starts the system manager thread.
        """
        logger.debug("Starting system monitor")
        self.running = True
        self.thread.start()

    def stop_monitor(self) -> None:
        """
        Stops the system manager thread.
        """
        logger.debug("Stopping system monitor")
        self.running = False
        self.thread.stop()
        self.thread.join()

    @property
    def engine_root_dir(self) -> str:
        """
        Gets the root location for the engine.
        """
        root = self.configuration.get("enfugue.engine.root", "~/.cache/enfugue")
        if root.startswith("~"):
            root = os.path.expanduser(root)
        root = os.path.realpath(root)
        check_make_directory(root)
        return root

    @property
    def engine_image_dir(self) -> str:
        """
        Gets the location for image result outputs.
        """
        directory = os.path.join(self.engine_root_dir, "images")
        check_make_directory(directory)
        return directory

    @property
    def engine_intermediate_dir(self) -> str:
        """
        Gets the location for image intermediate outputs.
        """
        directory = os.path.join(self.engine_root_dir, "intermediates")
        check_make_directory(directory)
        return directory

    @property
    def engine_models_dir(self) -> str:
        """
        Gets the location for models.
        """
        directory = os.path.join(self.engine_root_dir, "models")
        check_make_directory(directory)
        return directory

    @property
    def engine_checkpoint_dir(self) -> str:
        """
        Returns the engine checkpoint location.
        """
        path = os.path.join(self.engine_root_dir, "checkpoint")
        check_make_directory(path)
        return path

    @property
    def default_model_ckpt(self) -> str:
        """
        Returns the location where the default model should be.
        """
        default_model_ckpt = os.path.basename(DEFAULT_MODEL)
        return os.path.join(self.engine_checkpoint_dir, default_model_ckpt)

    @property
    def default_inpaint_ckpt(self) -> str:
        """
        Returns the location where the default inpaint model should be.
        """
        default_inpaint_ckpt = os.path.basename(DEFAULT_INPAINTING_MODEL)
        return os.path.join(self.engine_checkpoint_dir, default_inpaint_ckpt)

    @property
    def pending_default_downloads(self) -> List[Tuple[str, str]]:
        """
        Gets default downloads that need to be started.
        """
        default_model_ckpt, default_inpaint_ckpt = (
            self.default_model_ckpt,
            self.default_inpaint_ckpt,
        )
        pending = []
        if not os.path.exists(default_model_ckpt) and not self.is_downloading(DEFAULT_MODEL):
            pending.append((DEFAULT_MODEL, default_model_ckpt))
        if not os.path.exists(default_inpaint_ckpt) and not self.is_downloading(
            DEFAULT_INPAINTING_MODEL
        ):
            pending.append((DEFAULT_INPAINTING_MODEL, default_inpaint_ckpt))
        return pending

    @property
    def active_default_downloads(self) -> List[Download]:
        """
        Gets default downloads that are currently underway.
        """
        return [
            download
            for download in self.active_downloads
            if download.src in [DEFAULT_MODEL, DEFAULT_INPAINTING_MODEL]
        ]

    @property
    def is_downloading_defaults(self) -> bool:
        """
        Returns true if there is an active download of a default model.
        We queue invocations until these are done, otherwise we might spawn
        a conflicting download process.
        """
        return len(self.active_default_downloads) > 0

    @property
    def max_queued_downloads(self) -> int:
        """
        The maximum number of downloads that can be queued.
        """
        return self.configuration.get("enfugue.downloads.queue", self.DEFAULT_MAX_QUEUED_DOWNLOADS)

    @property
    def max_concurrent_downloads(self) -> int:
        """
        The maximum number of downloads that can go at once.
        """
        return self.configuration.get(
            "enfugue.downloads.concurrent", self.DEFAULT_MAX_CONCURRENT_DOWNLOADS
        )

    @property
    def max_queued_invocations(self) -> int:
        """
        The maximum number of invocations that can be queued.
        """
        return self.configuration.get("enfugue.queue", self.DEFAULT_MAX_QUEUED_INVOCATIONS)

    @property
    def max_intermediate_age(self) -> int:
        """
        Gets the maximum age of an intermediate image (in seconds)
        After this time, it is eligible for removal.
        """
        return self.configuration.get("enfugue.intermediates.age", 3600)

    @property
    def max_timing_cache_age(self) -> int:
        """
        Gets the maxium age for a TensorRT timing cache, in seconds
        """
        return self.configuration.get("enfugue.tensorrt.age", 60 * 60 * 24 * 30)

    @property
    def active_downloads(self) -> List[Download]:
        """
        Gets a list of active downloads
        """
        return [
            download
            for download_list in self.downloads.values()
            for download in download_list
            if not download.complete
        ]

    @property
    def remaining_concurrent_downloads(self) -> int:
        """
        Gets the remaining number of download slots.
        """
        return self.max_concurrent_downloads - len(self.active_downloads)

    @property
    def can_start_download(self) -> bool:
        """
        Returns true if there are any download slots remaining.
        """
        return self.remaining_concurrent_downloads > 0

    @property
    def can_queue_download(self) -> bool:
        """
        Retrurns true if there are any download queue slots remaining.
        """
        return len(self.download_queue) < self.max_queued_downloads

    @property
    def can_invoke(self) -> bool:
        """
        Returns true if there is no active invocation.
        """
        return self.active_invocation is None and not self.is_downloading_defaults

    @property
    def can_queue_invocation(self) -> bool:
        """
        Returns true if there are open invocation queue spots.
        """
        return len(self.invocation_queue) < self.max_queued_invocations

    @property
    def engine_status(self) -> str:
        """
        Gets a textual description of the engine status.
        """
        if not self.can_invoke:
            return "busy"
        return "ready" if self.engine.keepalive() else "idle"

    @property
    def status(self) -> SystemManagerStatusDict:
        """
        Gets the status of the manager in a dict.
        """
        return {
            "downloads": {
                "active": len(self.active_downloads),
                "queued": len(self.download_queue),
                "total": sum([len(download_list) for download_list in self.downloads.values()]),
            },
            "invocations": {
                "active": not self.can_invoke,
                "queued": len(self.invocation_queue),
                "total": sum(
                    [len(invocation_list) for invocation_list in self.invocations.values()]
                ),
            },
        }

    def is_downloading(self, url: str) -> bool:
        """
        Returns true if there is already an active download process for a URL.
        """
        return len([download for download in self.active_downloads if download.src == url]) > 0

    def download(
        self,
        user_id: int,
        url: str,
        destination: str,
        headers: Dict[str, str] = {},
        parameters: Dict[str, Any] = {},
    ) -> Download:
        """
        Starts a download process.
        """
        if self.is_downloading(url):
            raise BadRequestError(f"Already downloading {url}")

        can_start = self.can_start_download
        can_queue = self.can_queue_download

        if not can_start and not can_queue:
            raise TooManyRequestsError()

        download = Download(url, destination, headers=headers, parameters=parameters, progress=True)
        if user_id not in self.downloads:
            self.downloads[user_id] = []
        self.downloads[user_id].append(download)
        if can_start:
            download.start()
        else:
            self.download_queue.append(download)

        return download

    def invoke(
        self,
        user_id: int,
        plan: DiffusionPlan,
        disable_intermediate_decoding: bool = False,
        **kwargs: Any,
    ) -> Invocation:
        """
        Starts an invocation.
        """
        can_start = self.can_invoke
        can_queue = self.can_queue_invocation

        if not can_start and not can_queue:
            raise TooManyRequestsError()

        if disable_intermediate_decoding:
            kwargs["decode_nth_intermediate"] = None

        invocation = Invocation(
            engine=self.engine,
            plan=plan,
            engine_image_dir=self.engine_image_dir,
            engine_intermediate_dir=self.engine_intermediate_dir,
            **kwargs,
        )  # TODO: Add config for safety

        if can_start:
            invocation.start()
            self.active_invocation = invocation
        else:
            self.invocation_queue.append(invocation)
        if user_id not in self.invocations:
            self.invocations[user_id] = []
        self.invocations[user_id].append(invocation)
        return invocation

    def get_downloads(self, user_id: int) -> List[Download]:
        """
        Gets downloads for a user.
        """
        return self.downloads.get(user_id, [])

    def get_invocations(self, user_id: int) -> List[Invocation]:
        """
        Gets invocations for a user.
        """
        return self.invocations.get(user_id, [])

    def stop_engine(self) -> None:
        """
        Stops the engine forcibly.
        """
        if self.active_invocation is not None:
            try:
                self.active_invocation.terminate()
                time.sleep(5)
            except Exception as ex:
                logger.info(f"Ignoring exception during invocation termination: {ex}")
            self.active_invocation = None
        self.engine.terminate_process()

    def clean_intermediates(self) -> None:
        """
        Cleans up intermediate files
        """
        reclaimed_bytes = 0
        for file_name in os.listdir(self.engine_intermediate_dir):
            file_path = os.path.join(self.engine_intermediate_dir, file_name)
            file_mod_time = datetime.datetime.fromtimestamp(os.path.getmtime(file_path))
            file_age = (datetime.datetime.now() - file_mod_time).total_seconds()
            if file_age > self.max_intermediate_age:
                logger.info(f"Removing intermediate {file_path}")
                reclaimed_bytes += os.path.getsize(file_path)
                os.remove(file_path)
        if reclaimed_bytes > 0:
            logger.info(f"Reclaimed {human_size(reclaimed_bytes)} from intermediates")

    def clean_models(self) -> None:
        """
        Cleans up models, engines, etc.
        """
        reclaimed_bytes = 0
        for model_name in os.listdir(self.engine_models_dir):
            model_path = os.path.join(self.engine_models_dir, model_name)
            model_tensorrt_dir = os.path.join(model_path, "tensorrt")
            if os.path.exists(model_tensorrt_dir):
                for stage in ["clip", "unet", "vae"]:
                    model_tensorrt_stage_dir = os.path.join(model_tensorrt_dir, stage)
                    if os.path.exists(model_tensorrt_stage_dir):
                        for stage_key in os.listdir(model_tensorrt_stage_dir):
                            stage_dir = os.path.join(model_tensorrt_stage_dir, stage_key)
                            stage_files = os.listdir(stage_dir)
                            to_remove = []

                            for file_name in stage_files:
                                file_path = os.path.join(stage_dir, file_name)
                                if file_name == "model.opt.onnx":
                                    if "engine.plan" in stage_files:
                                        to_remove.append(file_path)
                                elif file_name == "model.onnx":
                                    if "model.opt.onnx" in stage_files:
                                        to_remove.append(file_path)
                                elif file_name == "timing_cache":
                                    file_mod_time = datetime.datetime.fromtimestamp(
                                        os.path.getmtime(file_path)
                                    )
                                    file_age = (
                                        datetime.datetime.now() - file_mod_time
                                    ).total_seconds()
                                    if (
                                        file_age > self.max_timing_cache_age
                                        or "engine.plan" in stage_files
                                    ):
                                        to_remove.append(file_path)
                                elif file_name != "engine.plan" and file_name != "metadata.json":
                                    to_remove.append(file_path)

                            for file_path in to_remove:
                                logger.info(f"Removing unneeded engine file {file_path}")
                                reclaimed_bytes += os.path.getsize(file_path)
                                os.remove(file_path)
        if reclaimed_bytes > 0:
            logger.info(f"Reclaimed {human_size(reclaimed_bytes)} from models")

    def do_periodic_tasks(self) -> None:
        """
        Looks at the queues and starts actions if necessary.
        """
        if not self.running:
            raise IOError("Manager stopped, periodic tasks should be ceased.")
        with self.lock:
            available_downloads = self.remaining_concurrent_downloads
            while available_downloads > 0:
                try:
                    next_download = self.download_queue.pop(0)
                    next_download.start()
                    available_downloads -= 1
                except IndexError:
                    break

            self.clean_intermediates()
            self.clean_models()

            if self.active_invocation is not None:
                if self.active_invocation.is_dangling:
                    logger.info("Active invocation appears to be dangling, terminating it.")
                    self.active_invocation.timeout()
                    time.sleep(5)
                elif (
                    self.active_invocation.results is None and self.active_invocation.error is None
                ):
                    try:
                        self.active_invocation.poll()
                    except IOError:
                        self.active_invocation = (
                            None  # results came in between checking and polling
                        )
                else:
                    self.active_invocation = None

            if self.active_invocation is None and not self.is_downloading_defaults:
                try:
                    self.active_invocation = self.invocation_queue.pop(0)
                    self.active_invocation.start()
                except IndexError:
                    for invocation_list in self.invocations.values():
                        for invocation in invocation_list:
                            if invocation.id is None:
                                logger.warning(
                                    "Unstarted invocation found that wasn't on invocation queue. Pushing it back on the queue."
                                )
                                self.invocation_queue.append(invocation)
                            elif invocation.is_dangling:
                                logger.warning("Dangling invocation found, timing it out.")
                                invocation.timeout()



import torch

from typing import Tuple, Dict, List, Union

from onnx.onnx_ml_pb2 import ModelProto as ONNXModel
from enfugue.diffusion.rt.optimizer import Optimizer


class BaseModel:
    """
    The BaseModel maps a diffusers model to an ONNX graph.
    """

    def __init__(
        self,
        model: torch.nn.Module,
        use_fp16: bool = False,
        device: str = "cuda",
        max_batch_size: int = 16,
        embedding_dim: int = 768,
        text_maxlen: int = 77,
    ) -> None:
        self.model = model
        self.name = "SDModel"
        self.use_fp16 = use_fp16
        self.device = device

        self.min_batch = 1
        self.max_batch = max_batch_size
        self.min_image_shape = 256  # min image resolution: 256x256
        self.max_image_shape = 1024  # max image resolution: 1024x1024
        self.min_latent_shape = self.min_image_shape // 8
        self.max_latent_shape = self.max_image_shape // 8

        self.embedding_dim = embedding_dim
        self.text_maxlen = text_maxlen

    def get_model(self) -> torch.nn.Module:
        """
        Gets the NN module.
        """
        return self.model

    def get_model_key(self) -> str:
        """
        Gets the key name of the model. Specific to each model.
        """
        raise NotImplementedError()

    def get_input_names(self) -> List[str]:
        """
        Gets the input names that will be passed into the graph.
        """
        raise NotImplementedError()

    def get_output_names(self) -> List[str]:
        """
        Gets the output names that will come from the graph.
        """
        raise NotImplementedError()

    def get_dynamic_axes(self) -> Dict[str, Dict[int, str]]:
        """
        Gets dynamic axes that can be shared.

        You should try your best to give matching names to all axes that
        could potentially be shared (i.e. they have the same dimensions.)

        For example, if you know the second and third axes of multiple
        input tensors will match the latent height/width, you could do something like this:
            return {
                "tensor1": {2: "H", 3: "W"},
                "tensor2": {2: "H", 3: "W"}
            }
        The actual value given to the axis is unimportant, it needs only be consistent.
        """
        raise NotImplementedError()

    def get_sample_input(
        self, batch_size: int, image_height: int, image_width: int
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, ...]]:
        """
        Gets sample input to go to the graph, which will be used to build the TRT engine
        """
        raise NotImplementedError()

    def get_input_profile(
        self,
        batch_size: int,
        image_height: int,
        image_width: int,
        static_batch: bool,
        static_shape: bool,
    ) -> Dict[str, List[Tuple[int, ...]]]:
        """
        Gets the min/opt/max dimensions for all input names
        """
        raise NotImplementedError()

    def get_shape_dict(
        self, batch_size: int, image_height: int, image_width: int
    ) -> Dict[str, Tuple[int, ...]]:
        """
        Gets the optimal dimensions for all inputs and outputs
        """
        raise NotImplementedError()

    def optimize(self, onnx_graph: ONNXModel) -> ONNXModel:
        """
        Runs the optimizer, returning the optimized version.
        """
        return Optimizer.run(onnx_graph)

    def check_dims(self, batch_size: int, image_height: int, image_width: int) -> Tuple[int, int]:
        """
        Checks dimensions.
        :raises AssertionError:
        """
        assert batch_size >= self.min_batch and batch_size <= self.max_batch
        assert image_height % 8 == 0 or image_width % 8 == 0
        latent_height = image_height // 8
        latent_width = image_width // 8
        assert latent_height >= self.min_latent_shape and latent_height <= self.max_latent_shape
        assert latent_width >= self.min_latent_shape and latent_width <= self.max_latent_shape
        return (latent_height, latent_width)

    def get_minmax_dims(
        self,
        batch_size: int,
        image_height: int,
        image_width: int,
        static_batch: bool,
        static_shape: bool,
    ) -> Tuple[
        Tuple[int, int],
        Tuple[int, int],
        Tuple[int, int],
        Tuple[int, int],
        Tuple[int, int],
    ]:
        """
        Gets min/max for:
            batch
            image_height
            image_width
            latent_height
            latent_width
        """
        min_batch = batch_size if static_batch else self.min_batch
        max_batch = batch_size if static_batch else self.max_batch
        latent_height = image_height // 8
        latent_width = image_width // 8
        min_image_height = image_height if static_shape else self.min_image_shape
        max_image_height = image_height if static_shape else self.max_image_shape
        min_image_width = image_width if static_shape else self.min_image_shape
        max_image_width = image_width if static_shape else self.max_image_shape
        min_latent_height = latent_height if static_shape else self.min_latent_shape
        max_latent_height = latent_height if static_shape else self.max_latent_shape
        min_latent_width = latent_width if static_shape else self.min_latent_shape
        max_latent_width = latent_width if static_shape else self.max_latent_shape
        return (
            (min_batch, max_batch),
            (min_image_height, max_image_height),
            (min_image_width, max_image_width),
            (min_latent_height, max_latent_height),
            (min_latent_width, max_latent_width),
        )



from enfugue.database.base import EnfugueObjectBase
from sqlalchemy import Column, String, Integer, Float


class DiffusionModel(EnfugueObjectBase):
    __tablename__ = "model"

    name = Column(String(256), primary_key=True)

    model = Column(String(256), nullable=False)
    size = Column(Integer, nullable=False)
    prompt = Column(String(256), nullable=False, default="")
    negative_prompt = Column(String(256), nullable=False, default="")


class DiffusionModelLora(EnfugueObjectBase):
    __tablename__ = "model_lora"

    diffusion_model_name = Column(
        DiffusionModel.ForeignKey("name", ondelete="CASCADE", onupdate="CASCADE"), primary_key=True
    )
    model = Column(String(256), primary_key=True)
    weight = Column(Float, default=1.0, nullable=False)

    diffusion_model = DiffusionModel.Relationship(backref="lora")


class DiffusionModelInversion(EnfugueObjectBase):
    __tablename__ = "model_inversion"

    diffusion_model_name = Column(
        DiffusionModel.ForeignKey("name", ondelete="CASCADE", onupdate="CASCADE"), primary_key=True
    )
    model = Column(String(256), primary_key=True)

    diffusion_model = DiffusionModel.Relationship(backref="inversion")



from typing import Any, Tuple, Iterable, Dict
from pibble.database.orm import ORM

__all__ = ["EnfugueConfiguration"]


class NoDefault:
    pass


class EnfugueConfiguration:
    """
    This class allows us to get and set items to the database using a dict syntax
    """

    def __init__(self, orm: ORM) -> None:
        self.orm = orm

    def get(self, key: str, default: Any = NoDefault) -> Any:
        """
        Gets an item, allow a default
        """
        try:
            return self[key]
        except KeyError:
            if default is NoDefault:
                raise
            return default

    def items(self) -> Iterable[Tuple[str, Any]]:
        """
        Iterates over the entire set of key-value pairs.
        """
        with self.orm.session() as session:
            for item in session.query(self.orm.ConfigurationItem).all():
                yield (item.configuration_key, item.configuration_value)

    def dict(self) -> Dict[str, Any]:
        """
        Gets all items as a dict.
        """
        return dict([item for item in self.items()])

    def __getitem__(self, key: str) -> Any:
        """
        Gets an item from the database
        """
        with self.orm.session() as session:
            item = (
                session.query(self.orm.ConfigurationItem)
                .filter(self.orm.ConfigurationItem.configuration_key == key)
                .one_or_none()
            )
            if item is None:
                raise KeyError(key)
            return item.configuration_value

    def __setitem__(self, key: str, value: Any) -> None:
        """
        Sets one item in the database
        """
        with self.orm.session() as session:
            item = (
                session.query(self.orm.ConfigurationItem)
                .filter(self.orm.ConfigurationItem.configuration_key == key)
                .one_or_none()
            )
            if item is None:
                session.add(
                    self.orm.ConfigurationItem(configuration_key=key, configuration_value=value)
                )
            else:
                item.configuration_value = value
            session.commit()



import os
import io
import enfugue

from webob import Request, Response

from typing import Any, Union, Dict

from lxml import etree as ET
from lxml.builder import E

from pibble.util.helpers import url_join
from pibble.util.imaging import contrast_color
from pibble.util.strings import snake_case, Serializer

from pibble.api.exceptions import NotFoundError
from pibble.api.server.webservice.template import TemplateServer

from pibble.api.middleware.database.orm import ORMMiddlewareBase

from pibble.ext.cms.middleware import CMSExtensionContextMiddleware
from pibble.ext.cms.server.extension import (
    CMSExtensionStaticExtension,
    CMSExtensionResolveStatementExtension,
    CMSExtensionResolveFunctionExtension,
)
from pibble.ext.user.server.base import (
    UserExtensionTemplateServer,
    UserExtensionTemplateHandlerRegistry,
    UserExtensionServerBase,
)
from pibble.ext.session.server.base import SessionExtensionServerBase

from enfugue.util import logger, get_local_static_directory
from enfugue.interface.helpers import (
    HTMLPropertiesHelperFunction,
    SerializeHelperFunction,
    SerializeHelperFilter,
    CheckResolveURLHelperFilter,
)
from enfugue.database import *


class EnfugueInterfaceServer(
    UserExtensionTemplateServer,
    CMSExtensionContextMiddleware,
    SessionExtensionServerBase,
):
    handlers = UserExtensionTemplateHandlerRegistry()

    def configure(self, **configuration: Any) -> None:
        """
        Intercept configure() to add the static directory.
        """
        production_root = get_local_static_directory()
        production_templates = os.path.join(production_root, "html")

        development_root = os.path.abspath(
            os.path.join(os.path.dirname(enfugue.__file__), "../../../src")
        )
        development_templates = os.path.join(development_root, "html")

        static_dirs = [production_root]
        template_dirs = []

        if os.path.exists(development_root):
            static_dirs.append(development_root)
            if os.path.exists(development_templates):
                template_dirs.append(development_templates)
        if os.path.exists(production_templates):
            template_dirs.append(production_templates)

        server = configuration.get("server", {})
        static_config = server.get("static", {})
        static_directories = static_config.get("directories", [])
        static_directories.extend(static_dirs)
        static_config["directories"] = static_directories
        server["static"] = static_config

        template_config = server.get("template", {})
        template_directories = template_config.get("directories", [])
        template_directories.extend(template_dirs)
        template_config["directories"] = template_directories
        server["template"] = template_config

        configuration["server"] = server
        super(EnfugueInterfaceServer, self).configure(**configuration)

    def on_configure(self):
        """
        The application interface extends the entire Enfugue database structure,
        with the exception of admin objects.
        """
        self.orm.extend_base(EnfugueObjectBase)
        self.templates.extend(
            CMSExtensionStaticExtension,
            CMSExtensionResolveStatementExtension,
            CMSExtensionResolveFunctionExtension,
            HTMLPropertiesHelperFunction,
            CheckResolveURLHelperFilter,
            SerializeHelperFunction,
            SerializeHelperFilter,
        )

    def prepare_context(self, request: Request, response: Response) -> Dict[str, Any]:
        """
        This is the function called by the TemplateServer to get the context available
        in templates.
        """
        context = {
            "current_path": request.path,
            "current_url": request.url,
            "canonical_url": url_join(self.configuration["server.cms.path.root"], request.path),
            "orm": self.orm,
            "database": self.database,
            "paths": self.configuration["server.cms.path"],
        }

        if hasattr(request, "token"):
            context["user"] = request.token.user
            context["admin"] = self.check_user_permission(request.token.user, "System", "update")

        return context

    @handlers.path("^/$")
    @handlers.methods("GET")
    @handlers.template("content-application.html.j2")
    @handlers.reverse("Application", "/")
    @handlers.secured()
    def root(self, request: Request, response: Response) -> Dict[str, Any]:
        """
        At the root, we display the application.

        If the user isn't logged in, they'll see a login there.
        """
        return {"title": "Application"}

    @handlers.bypass(
        SessionExtensionServerBase,
        UserExtensionTemplateServer,
        TemplateServer,
        UserExtensionServerBase,
        ORMMiddlewareBase,
    )  # Bypass token checking and ORM instantiation
    @handlers.methods("GET")
    @handlers.cache(86400)
    @handlers.path("^/manifest.json$")
    def get_manifest(self, request: Request, response: Response) -> None:
        """
        Gets the browser manifest.
        """
        app_name = self.configuration.get("server.cms.name", "Pibble App Template")
        app_color = self.configuration.get("server.cms.context.base.meta.theme-color", "#ffffff")

        response.content_type = "application/json"
        response.json = {
            "name": app_name,
            "short_name": self.configuration.get("server.cms.short_name", snake_case(app_name)),
            "icons": [
                {
                    "src": url_join(
                        self.configuration["server.cms.path.root"],
                        self.configuration["server.cms.context.base.android.icon.href"],
                    ),
                    "sizes": self.configuration["server.cms.context.base.android.icon.size"],
                    "type": self.configuration["server.cms.context.base.android.icon.type"],
                }
            ],
            "theme_color": self.configuration.get(
                "server.cms.context.base.meta.theme-color", "#ffffff"
            ),
            "background_color": "#000000" if contrast_color(app_color) == "black" else "#ffffff",
            "display": "standalone",
        }

    @handlers.bypass(
        SessionExtensionServerBase,
        UserExtensionTemplateServer,
        TemplateServer,
        UserExtensionServerBase,
        ORMMiddlewareBase,
    )  # Bypass token checking and ORM instantiation
    @handlers.methods("GET")
    @handlers.cache(86400)
    @handlers.path("^/browserconfig.xml$")
    def get_browserconfig(self, request: Request, response: Response) -> None:
        """
        Gets the 'browserconfig' - an xml config for windows shortcuts.
        """
        app_color = self.configuration.get("server.cms.context.base.meta.theme-color", "#ffffff")
        response.content_type = "application/xml"
        response.body = ET.tostring(
            E.browserconfig(
                E.msapplication(
                    E.tile(
                        E(
                            "square{0}logo".format(
                                self.configuration["server.cms.context.base.windows.icon.size"]
                            ),
                            src=url_join(
                                self.configuration["server.cms.path.root"],
                                self.configuration["server.cms.context.base.windows.icon.href"],
                            ),
                        ),
                        E.TileColor(app_color),
                    )
                )
            ),
            xml_declaration=True,
            encoding="utf-8",
        )

    @handlers.bypass(
        SessionExtensionServerBase,
        UserExtensionTemplateServer,
        TemplateServer,
        UserExtensionServerBase,
        ORMMiddlewareBase,
    )  # Bypass token checking and ORM instantiation
    @handlers.methods("GET")
    @handlers.cache(86400)
    @handlers.path("^/favicon.ico$")
    def get_root_favicon(self, request: Request, response: Response) -> None:
        """
        Gets the root favicon, which browsers will sometimes request before
        the page has completed loading (which would have told them the .ico is
        on the static path)
        """
        self.redirect(
            response,
            url_join(
                self.configuration["server.cms.path.root"],
                self.configuration["server.cms.context.base.links"]["shortcut icon"][0]["href"],
            ),
        )

    @handlers.bypass(
        SessionExtensionServerBase,
        UserExtensionTemplateServer,
        TemplateServer,
        UserExtensionServerBase,
        ORMMiddlewareBase,
    )  # Bypass token checking and ORM instantiation
    @handlers.reverse("static", "/static/{path}")
    @handlers.methods("GET")
    @handlers.path("/static/(?P<path>\S+)")
    @handlers.cache()
    @handlers.download()
    def static(self, request: Request, response: Response, path: str) -> Union[str, io.BytesIO]:
        directories = self.configuration.get("server.static.directories", [])
        if path.endswith("/index.autogenerated.mjs"):
            # Create autogenerated file index for JS directories.
            for root in directories:
                dir_path = os.path.dirname(os.path.join(root, path))
                if os.path.exists(dir_path) and os.path.isdir(dir_path):
                    js_files = [
                        filename for filename in os.listdir(dir_path) if filename.endswith("js")
                    ]
                    if len(js_files) > 0:
                        index = "const Index = {0}; export {{ Index }};".format(
                            Serializer.serialize(list(sorted(js_files)))
                        )
                        response.content_type = "application/javascript"
                        return io.BytesIO(index.encode("utf-8"))
            raise NotFoundError(f"No javascript files found at path {path[:-23]}")
        for root in directories:
            root_path = os.path.join(root, path)
            if os.path.exists(root_path) and os.path.isfile(root_path):
                return root_path
        logger.debug(f"Couldn't find path {path}, tried {directories}")
        raise NotFoundError(f"No file at {path}")

    @handlers.errors(401)
    def handle_authentication_error(self, request: Request, response: Response) -> None:
        """
        Determine what to do when authentication errors occur, then do it.
        """
        if self.configuration.get("enfugue.noauth", False):
            # Bypass login
            self.bypass_login(request, response)
            logger.critical(":BYOP")
            self.handle_request(request, response)  # type: ignore
        else:
            self.redirect(response, "/login", 302)

    @handlers.cache(86400)
    @handlers.template("content-login.html.j2")
    @handlers.methods("GET")
    @handlers.path("^/login$")
    def view_login(self, request: Request, response: Response) -> Dict[str, Any]:
        """
        The GET() handler for logging in, i.e, the landing page when a user isn't logged in.
        """
        return {"title": "Login"}

    @handlers.methods("POST")
    @handlers.template("content-login.html.j2")
    @handlers.path("^/login$")
    @handlers.reverse("Login", "/login")
    def process_login(self, request, response) -> Dict[str, Any]:
        """
        The POST() handler for logging in.
        """
        context = {"title": "Login"}
        try:
            token = self.login(request, response)
        except Exception as ex:
            context["error"] = str(ex)
            context["username"] = request.POST["username"]
        else:
            self.redirect(
                response,
                request.POST["redirect"]
                if "redirect" in request.POST
                else url_join(
                    self.configuration["server.cms.path.root"],
                    self.resolve("Application", controller="data"),
                ),
            )
        return context

    @handlers.secured()
    @handlers.methods("GET")
    @handlers.reverse("Logout", "/logout")
    @handlers.path("^/logout$")
    def process_logout(self, request: Request, response: Response) -> None:
        """
        Logout() is a GET() handler that just calls the UserExtension parent method.
        """
        context = {"title": "Logout"}
        self.logout(request, response)
        self.redirect(
            response,
            url_join(self.configuration["server.cms.path.root"], self.resolve("Login")),
        )



from __future__ import annotations

import os

from typing import Iterator, Callable

from pibble.api.server.webservice.jsonapi import JSONWebServiceAPIServer
from enfugue.util import check_make_directory
from enfugue.api.invocations import Invocation
from enfugue.api.downloads import Download
from enfugue.api.manager import SystemManager

__all__ = ["EnfugueAPIControllerBase"]


class EnfugueAPIControllerBase(JSONWebServiceAPIServer):
    invoke: Callable[..., Invocation]
    download: Callable[[int, str, str], Download]
    manager: SystemManager

    @property
    def safe(self) -> bool:
        """
        Returns whether or not to use safety filters.
        """
        return self.configuration.get("enfugue.safe", True)

    @property
    def engine_root(self) -> str:
        """
        Returns the engine root location.
        """
        root = self.configuration.get("enfugue.engine.root", "~/.cache/enfugue")
        if root.startswith("~"):
            root = os.path.expanduser(root)
        root = os.path.realpath(root)
        check_make_directory(root)
        return root

    @staticmethod
    def enumerate() -> Iterator[EnfugueAPIControllerBase]:
        """
        Iterates through declared subcontroller classes.
        """
        for cls in EnfugueAPIControllerBase.__subclasses__():
            yield cls



import math
import io
import PIL
import PIL.Image

from typing import Optional, Literal

__all__ = ["fit_image", "feather_mask", "remove_background"]


def fit_image(
    image: PIL.Image.Image,
    width: int,
    height: int,
    fit: Optional[Literal["actual", "stretch", "cover", "contain"]] = None,
    anchor: Optional[
        Literal[
            "top-left",
            "top-center",
            "top-right",
            "center-left",
            "center-center",
            "center-right",
            "bottom-left",
            "bottom-center",
            "bottom-right",
        ]
    ] = None,
) -> PIL.Image.Image:
    """
    Given an image of unknown size, make it a known size with optional fit parameters.
    """
    if fit is None or fit == "actual":
        left, top = 0, 0
        crop_left, crop_top = 0, 0
        image_width, image_height = image.size

        if anchor is not None:
            top_part, left_part = anchor.split("-")

            if top_part == "center":
                top = height // 2 - image_height // 2
            elif top_part == "bottom":
                top = height - image_height

            if left_part == "center":
                left = width // 2 - image_width // 2
            elif left_part == "right":
                left = width - image_width

        blank_image = PIL.Image.new("RGBA", (width, height), (0, 0, 0, 0))
        blank_image.paste(image, (left, top))

        return blank_image
    elif fit == "contain":
        image_width, image_height = image.size
        width_ratio, height_ratio = width / image_width, height / image_height
        horizontal_image_width, horizontal_image_height = int(image_width * width_ratio), int(
            image_height * width_ratio
        )
        vertical_image_width, vertical_image_height = int(image_width * height_ratio), int(
            image_height * height_ratio
        )
        top, left = 0, 0
        direction = None
        if width >= horizontal_image_width and height >= horizontal_image_height:
            input_image = image.resize((horizontal_image_width, horizontal_image_height))
            if anchor is not None:
                top_part, _ = anchor.split("-")
                if top_part == "center":
                    top = height // 2 - horizontal_image_height // 2
                elif top_part == "bottom":
                    top = height - horizontal_image_height
        elif width >= vertical_image_width and height >= vertical_image_height:
            input_image = image.resize((vertical_image_width, vertical_image_height))
            if anchor is not None:
                _, left_part = anchor.split("-")
                if left_part == "center":
                    left = width // 2 - vertical_image_width // 2
                elif left_part == "right":
                    left = width - vertical_image_width
        blank_image = PIL.Image.new("RGBA", (width, height))
        blank_image.paste(input_image, (left, top))

        return blank_image
    elif fit == "cover":
        image_width, image_height = image.size
        width_ratio, height_ratio = width / image_width, height / image_height
        horizontal_image_width, horizontal_image_height = math.ceil(
            image_width * width_ratio
        ), math.ceil(image_height * width_ratio)
        vertical_image_width, vertical_image_height = math.ceil(
            image_width * height_ratio
        ), math.ceil(image_height * height_ratio)
        top, left = 0, 0
        direction = None
        if width <= horizontal_image_width and height <= horizontal_image_height:
            input_image = image.resize((horizontal_image_width, horizontal_image_height))
            if anchor is not None:
                top_part, _ = anchor.split("-")
                if top_part == "center":
                    top = height // 2 - horizontal_image_height // 2
                elif top_part == "bottom":
                    top = height - horizontal_image_height
        elif width <= vertical_image_width and height <= vertical_image_height:
            input_image = image.resize((vertical_image_width, vertical_image_height))
            if anchor is not None:
                _, left_part = anchor.split("-")
                if left_part == "center":
                    left = width // 2 - vertical_image_width // 2
                elif left_part == "right":
                    left = width - vertical_image_width
        else:
            input_image = image.resize((width, height))  # We're probably off by a pixel
        blank_image = PIL.Image.new("RGBA", (width, height))
        blank_image.paste(input_image, (left, top))

        return blank_image
    elif fit == "stretch":
        return image.resize((width, height)).convert("RGBA")
    else:
        raise ValueError(f"Unknown fit {fit}")


def feather_mask(image: PIL.Image.Image) -> PIL.Image.Image:
    """
    Given an image, create a feathered binarized mask by 'growing' the black/white pixel sections.
    """
    width, height = image.size

    mask = image.convert("L")
    feathered = mask.copy()

    for x in range(width):
        for y in range(height):
            if mask.getpixel((x, y)) == (0):
                for dx, dy in [(1, 0), (-1, 0), (0, 1), (0, -1)]:
                    nx, ny = x + dx, y + dy
                    if 0 <= nx < width and 0 <= ny < height and mask.getpixel((nx, ny)) == 255:
                        feathered.putpixel((x, y), (255))
                        break

    return feathered


def remove_background(image: PIL.Image.Image) -> PIL.Image.Image:
    """
    Remove the background from an image.
    """
    import backgroundremover.utilities

    # We have to import this in this order for backgroundremover to work
    backgroundremover.utilities  # silence importchecker
    import backgroundremover.bg

    buf = io.BytesIO()
    image.save(buf, "PNG")
    return PIL.Image.open(io.BytesIO(backgroundremover.bg.remove(buf.getvalue())))



from typing import Any

from pibble.util.strings import Serializer, decode
from pibble.util.helpers import url_join
from pibble.api.configuration import APIConfiguration
from pibble.api.server.webservice.template.extensions import (
    FunctionExtensionBase,
    FilterExtensionBase,
)

__all__ = [
    "HTMLPropertiesHelperFunction",
    "SerializeHelperFunction",
    "SerializeHelperFilter",
    "CheckResolveURLHelperFilter",
]


def check_url(configuration: APIConfiguration, value: Any) -> Any:
    """
    For configuration values that start with '/', prepend the CMS root.
    For configuration values that start with '/static', prepend the static root.
    """
    if isinstance(value, str) and value.startswith("/"):
        if value.startswith("/static/"):
            static_path = str(
                configuration.get("server.cms.path.static", configuration["server.cms.path.root"])
            )
            if static_path.endswith("static") or static_path.endswith("static/"):
                value = value[len("/static") :]
            return url_join(static_path, value)
        else:
            return url_join(str(configuration["server.cms.path.root"]), decode(value))
    return value


class HTMLPropertiesHelperFunction(FunctionExtensionBase):
    """
    Allows the transformation of a dictionary of properties into HTML key=value syntax.

    Callable in templates via {{ html_properties({"foo": "bar"}) }}

    >>> from enfugue.interface.helpers import HTMLPropertiesHelperFunction
    >>> from pibble.api.configuration import APIConfiguration
    >>> config = APIConfiguration(server = {'cms':{'path':{'root':'http://www.example.com'}}})
    >>> self = type('', (object,), {'getConfiguration': lambda: config})
    >>> HTMLPropertiesHelperFunction.__call__(self, {"id": "something", "name": "somethingelse"})
    'id="something" name="somethingelse"'
    >>> HTMLPropertiesHelperFunction.__call__(self, {"src": "/static/img.jpg"})
    'src="http://www.example.com/static/img.jpg"'
    """

    name = "html_properties"

    def __call__(self, property_dict: dict) -> str:
        for property_name in property_dict:
            property_dict[property_name] = check_url(
                self.getConfiguration(), property_dict[property_name]
            )

        return " ".join(
            [
                '{0}="{1}"'.format(key, property_dict[key])
                if property_dict[key] is not None
                else key
                for key in property_dict
            ]
        )


class SerializeHelperFunction(FunctionExtensionBase):
    """
    Allows calling serialize in templates.

    Example usage is {{ serialize({"foo: "bar"}) }}.

    >>> from enfugue.interface.helpers import SerializeHelperFunction
    >>> SerializeHelperFunction.__call__(None, {"id": "something", "name": "somethingelse"})
    '{"id": "something", "name": "somethingelse"}'
    >>> SerializeHelperFunction.__call__(None, {"src": None, "length": 4})
    '{"src": null, "length": 4}'
    """

    name = "serialize"

    def __call__(self, var: Any, **kwargs: Any):
        return Serializer.serialize(var, **kwargs)


class CheckResolveURLHelperFilter(FilterExtensionBase):
    """
    Allows calling check_resolve_url() in the template layer.

    Example usage is {{ img | check_resolve_url }}.

    >>> from enfugue.interface.helpers import CheckResolveURLHelperFilter
    >>> from pibble.api.configuration import APIConfiguration
    >>> config = APIConfiguration(server = {'cms':{'path':{'root':'http://www.example.com'}}})
    >>> self = type('', (object,), {'getConfiguration': lambda: config})
    >>> CheckResolveURLHelperFilter.__call__(self, "something")
    'something'
    >>> CheckResolveURLHelperFilter.__call__(self, 4)
    4
    >>> CheckResolveURLHelperFilter.__call__(self, "/static/img.jpg")
    'http://www.example.com/static/img.jpg'
    """

    name = "check_resolve_url"

    def __call__(self, var):
        return check_url(self.getConfiguration(), var)


class SerializeHelperFilter(FilterExtensionBase):
    """
    Allows using serialize as a filter in templates.

    Example usage is {{ {"foo: "bar"} | serialize }}.
    """

    name = "serialize"

    def __call__(self, var: Any):
        return Serializer.serialize(var)


__all__ = [
    "HTMLPropertiesHelperFunction",
    "CheckResolveURLHelperFilter",
    "SerializeHelperFunction",
    "SerializeHelperFilter",
]



from enfugue.api.controller.base import EnfugueAPIControllerBase
from enfugue.api.controller.downloads import EnfugueAPIDownloadsController
from enfugue.api.controller.invocation import EnfugueAPIInvocationController
from enfugue.api.controller.models import EnfugueAPIModelsController
from enfugue.api.controller.system import EnfugueAPISystemController

EnfugueAPIControllerBase, EnfugueAPISystemController, EnfugueAPIDownloadsController, EnfugueAPIInvocationController, EnfugueAPIModelsController

__all__ = [
    "EnfugueAPIControllerBase",
    "EnfugueAPISystemController",
    "EnfugueAPIDownloadsController",
    "EnfugueAPIInvocationController",
    "EnfugueAPIModelsController",
]



from enfugue.diffusion.rt.model.base import BaseModel
from enfugue.diffusion.rt.model.clip import CLIP
from enfugue.diffusion.rt.model.unet import UNet
from enfugue.diffusion.rt.model.controlledunet import ControlledUNet
from enfugue.diffusion.rt.model.controlnet import ControlNet
from enfugue.diffusion.rt.model.vae import VAE

BaseModel, CLIP, UNet, ControlledUNet, ControlNet, VAE

__all__ = ["BaseModel", "CLIP", "UNet", "ControlledUNet", "ControlNet", "VAE"]



import os
import base64
import datetime
import requests

from typing import List

from enfugue.util.log import logger
from enfugue.util.installation import get_local_config_directory

MAX_SIGNATURE_AGE = 60 * 60 * 24 * 30


__all__ = ["get_signature"]


def get_signature() -> List[str]:
    """
    Gets necessary signing details for the webserver
    """
    from enfugue.util.security import decrypt  # type: ignore[attr-defined]

    local = os.path.join(get_local_config_directory(), "payload.txt")
    contents = None

    if os.path.exists(local):
        saved = datetime.datetime.fromtimestamp(os.path.getmtime(local))
        if (datetime.datetime.now() - saved).total_seconds() > MAX_SIGNATURE_AGE:
            logger.info("Expiring local signatures.")
            os.remove(local)
        else:
            contents = open(local, "rb").read()
    if contents is None:
        logger.info("Fetching remote signatures.")
        contents = requests.get("https://cdn.enfugue.ai/payload.txt").content
        open(local, "wb").write(contents)

    return [
        base64.b64decode(decrypt(chunk)).decode("utf-8")
        for chunk in base64.b64decode(contents).decode("utf-8").split(":")
    ]
